
/* $Id: util.c,v 3.2 1998/03/13 16:02:57 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#include "flyme.h"
#include "flyme_glob.h"


HELI
create_heli()
{
	HELI tmp;

	tmp = (HELI) malloc( sizeof( struct heli_struct ) );
	memcheck( tmp, "Helicopter Structure" );

	tmp->win = (WIN) NULL;

	tmp->hid = -1;

	tmp->steered = -1;

	tmp->shield_str = -1;
	tmp->shielded = -1;
	tmp->last_shield_str = -1;

	tmp->score = -1;

	tmp->flash = 0;

	tmp->updated = -1;

	tmp->screen_x = -1;
	tmp->screen_y = -1;

	tmp->region_x = -1;
	tmp->region_y = -1;

	tmp->heading = -1.0;

	tmp->vel_angle = -1.0;

	tmp->velocity = -1.0;

	tmp->radar_range = -1.0;

	tmp->rotor_posn = -1;
	tmp->rotor_id = -1;
	tmp->body_id = -1;
	tmp->tail_id = -1;
	tmp->radar_id = -1;
	tmp->shield_id = -1;

	tmp->next = (HELI) NULL;

	return( tmp );
}


void
free_heli( ptr )
HELI *ptr;
{
	HELI H;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	H = *ptr;

	H->win = (WIN) NULL;

	H->hid = -1;

	H->steered = -1;

	H->shield_str = -1;
	H->shielded = -1;
	H->last_shield_str = -1;

	H->score = -1;

	H->flash = -1;

	H->updated = -1;

	H->screen_x = -1;
	H->screen_y = -1;

	H->region_x = -1;
	H->region_y = -1;

	H->heading = -1.0;

	H->vel_angle = -1.0;

	H->velocity = -1.0;

	H->radar_range = -1.0;

	H->rotor_posn = -1;
	H->rotor_id = -1;
	H->body_id = -1;
	H->tail_id = -1;
	H->radar_id = -1;
	H->shield_id = -1;

	H->next = (HELI) NULL;

	free( H );

	*ptr = (HELI) NULL;
}


HELI
get_heli_hid( W, hid )
WIN W;
int hid;
{
	HELI H;

	H = W->helis;

	while ( H != NULL )
	{
		if ( H->hid == hid )
			return( H );

		H = H->next;
	}

	return( (HELI) NULL );
}


HELI
closest_heli( W, x, y, not_me, close_dist )
WIN W;
int x, y;
HELI not_me;
double *close_dist;
{
	HELI Hclose;
	HELI H;

	double closest;
	double dist;

	Hclose = (HELI) NULL;

	closest = -1.0;

	H = W->helis;

	while ( H != NULL )
	{
		if ( not_me == NULL || H != not_me )
		{
			dist = distance( (double) H->screen_x, (double) H->screen_y,
				(double) x, (double) y );
		
			if ( dist < closest || closest == -1.0 )
			{
				Hclose = H;

				closest = dist;
			}
		}

		H = H->next;
	}

	if ( close_dist != NULL )
		*close_dist = closest;

	return( Hclose );
}


BULLET
create_bullet()
{
	BULLET tmp;

	tmp = (BULLET) malloc( sizeof( struct bullet_struct ) );
	memcheck( tmp, "Bullet Structure" );

	tmp->win = (WIN) NULL;

	tmp->bid = -1;

	tmp->updated = -1;

	tmp->screen_x = -1;
	tmp->screen_y = -1;

	tmp->region_x = -1;
	tmp->region_y = -1;

	tmp->vel_angle = -1.0;

	tmp->bullet_id = -1;

	tmp->next = (BULLET) NULL;

	return( tmp );
}


void
free_bullet( ptr )
BULLET *ptr;
{
	BULLET B;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	B = *ptr;

	B->win = (WIN) NULL;

	B->bid = -1;

	B->updated = -1;

	B->screen_x = -1;
	B->screen_y = -1;

	B->region_x = -1;
	B->region_y = -1;

	B->vel_angle = -1.0;

	B->bullet_id = -1;

	B->next = (BULLET) NULL;

	free( B );

	*ptr = (BULLET) NULL;
}


BULLET
get_bullet_bid( W, bid )
WIN W;
int bid;
{
	BULLET B;

	B = W->bullets;

	while ( B != NULL )
	{
		if ( B->bid == bid )
			return( B );

		B = B->next;
	}

	return( (BULLET) NULL );
}


EXPLOSION
create_explosion()
{
	EXPLOSION tmp;

	tmp = (EXPLOSION) malloc( sizeof( struct explosion_struct ) );
	memcheck( tmp, "Explosion Structure" );

	tmp->win = (WIN) NULL;

	tmp->eid = -1;

	tmp->updated = -1;

	tmp->screen_x = -1;
	tmp->screen_y = -1;

	tmp->region_x = -1;
	tmp->region_y = -1;

	tmp->angle = -1.0;

	tmp->strength = -1;

	tmp->radius = -1;

	tmp->explosion_id = -1;

	tmp->next = (EXPLOSION) NULL;

	return( tmp );
}


void
free_explosion( ptr )
EXPLOSION *ptr;
{
	EXPLOSION E;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	E = *ptr;

	E->win = (WIN) NULL;

	E->eid = -1;

	E->updated = -1;

	E->screen_x = -1;
	E->screen_y = -1;

	E->region_x = -1;
	E->region_y = -1;

	E->angle = -1.0;

	E->strength = -1;

	E->radius = -1;

	E->explosion_id = -1;

	E->next = (EXPLOSION) NULL;

	free( E );

	*ptr = (EXPLOSION) NULL;
}


EXPLOSION
get_explosion_eid( W, eid )
WIN W;
int eid;
{
	EXPLOSION E;

	E = W->explosions;

	while ( E != NULL )
	{
		if ( E->eid == eid )
			return( E );

		E = E->next;
	}

	return( (EXPLOSION) NULL );
}


EXPLOSION
closest_explosion( W, x, y, close_dist )
WIN W;
int x, y;
double *close_dist;
{
	EXPLOSION Eclose;
	EXPLOSION E;

	double closest;
	double dist;

	Eclose = (EXPLOSION) NULL;

	closest = -1.0;

	E = W->explosions;

	while ( E != NULL )
	{
		dist = distance( (double) E->screen_x, (double) E->screen_y,
			(double) x, (double) y );
	
		if ( dist < closest || closest == -1.0 )
		{
			Eclose = E;

			closest = dist;
		}

		E = E->next;
	}

	if ( close_dist != NULL )
		*close_dist = closest;

	return( Eclose );
}


GRID
create_grid()
{
	GRID tmp;

	tmp = (GRID) malloc( sizeof( struct grid_struct ) );
	memcheck( tmp, "Grid Line Structure" );

	tmp->id = -1;

	tmp->coord = -1;

	tmp->next = (GRID) NULL;

	return( tmp );
}


void
free_grid( ptr )
GRID *ptr;
{
	GRID G;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	G = *ptr;

	G->id = -1;

	G->coord = -1;

	G->next = (GRID) NULL;

	free( G );

	*ptr = (GRID) NULL;
}


void
free_grid_list( W, ptr )
WIN W;
GRID *ptr;
{
	GRID Gnext;
	GRID G;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	G = *ptr;

	*ptr = (GRID) NULL;

	while ( G != NULL )
	{
		if ( G->id != -1 )
			DELETE_ITEM( interp, W->canvas, G->id );
		
		Gnext = G->next;

		free_grid( &G );

		G = Gnext;
	}
}


WIN
create_win()
{
	WIN tmp;

	tmp = (WIN) malloc( sizeof( struct win_struct ) );
	memcheck( tmp, "Window Display Object" );

	tmp->canvas = (char *) NULL;

	tmp->htvar = (char *) NULL;
	tmp->wtvar = (char *) NULL;

	tmp->scale = -1.0;

	tmp->csx = tmp->csy = -1;

	stv_init_region( tmp->region );

	tmp->xoff = tmp->yoff = -1;

	tmp->helis = (HELI) NULL;

	tmp->bullets = (BULLET) NULL;

	tmp->explosions = (EXPLOSION) NULL;

	tmp->border_id = -1;

	tmp->xgrid = tmp->ygrid = (GRID) NULL;

	return( tmp );
}


void
free_win( ptr )
WIN *ptr;
{
	WIN W;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	W = *ptr;

	W->canvas = (char *) NULL;

	W->htvar = (char *) NULL;
	W->wtvar = (char *) NULL;

	W->scale = -1.0;

	W->csx = W->csy = -1;

	stv_init_region( W->region );

	W->xoff = W->yoff = -1;

	W->helis = (HELI) NULL;

	W->bullets = (BULLET) NULL;

	W->explosions = (EXPLOSION) NULL;

	W->border_id = -1;

	if ( W->xgrid != NULL )
		free_grid_list( W, &(W->xgrid) );

	if ( W->ygrid != NULL )
		free_grid_list( W, &(W->ygrid) );

	free( W );

	*ptr = (WIN) NULL;
}


REGION_STACK
create_region_stack()
{
	REGION_STACK tmp;

	int i;

	tmp = (REGION_STACK) malloc( sizeof( struct region_stack_struct ) );
	memcheck( tmp, "Region Stack Entry" );

	for ( i=0 ; i < STV_REGION_SIZE( STV_MAX_DIM ) ; i++ )
		tmp->region[i] = -1;
	
	tmp->next = (REGION_STACK) NULL;

	return( tmp );
}


void
free_region_stack( ptr )
REGION_STACK *ptr;
{
	REGION_STACK RS;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	RS = *ptr;

	for ( i=0 ; i < STV_REGION_SIZE( STV_MAX_DIM ) ; i++ )
		RS->region[i] = -1;
	
	RS->next = (REGION_STACK) NULL;

	free( RS );

	*ptr = (REGION_STACK) NULL;
}


void
push_region( R )
STV_REGION R;
{
	REGION_STACK RS;

	RS = create_region_stack();

	stv_copy_region( RS->region, R );

	RS->next = REGIONS;

	REGIONS = RS;
}


int
pop_region( R )
STV_REGION R;
{
	REGION_STACK RS;

	if ( REGIONS != NULL )
	{
		stv_copy_region( R, REGIONS->region );

		RS = REGIONS->next;

		free_region_stack( &REGIONS );

		REGIONS = RS;

		return( stvTrue );
	}

	return( stvFalse );
}


char *
center_str( str, width )
char *str;
int width;
{
	char *tmp;

	int i, j;
	int len;

	tmp = (char *) malloc( (STV_SIZE_T) (width + 1) * sizeof(char) );
	memcheck( tmp, "Center String" );

	len = strlen( str );

	if ( len > width )
	{
		strcpy( tmp, (char *) ( str + ( len - width ) ) );

		tmp[0] = '<';
	}

	else
	{
		j = ( width - len ) / 2;

		for ( i=0 ; i < j ; i++ )
			tmp[i] = ' ';

		strcpy( (char *) ( tmp + j ), str );

		for ( i=(j + len) ; i < width ; i++ )
			tmp[i] = ' ';

		tmp[width] = '\0';
	}

	return( tmp );
}


char *
pad_num( num, max )
long num;
long max;
{
	char tmp[1024];

	char *ptr;
	char *str;

	int nd;
	int nz;
	int i;

	sprintf( tmp, "%ld", num );

	nd = strlen( tmp );

	nz = max - nd;

	str = (char *) malloc( (STV_SIZE_T) (max + 1) * sizeof(char) );
	memcheck( str, "Numerical Pad String" );

	ptr = str;

	for ( i=0 ; i < nz ; i++ )
		*ptr++ = '0';

	sprintf( (char *) (str + nz), "%ld", num );

	return( str );
}


char *
trunc_str( str, len, mark_incl )
char *str;
int len;
int mark_incl;
{
	char tmp[1024];

	char *ptr;

	int slen;

	sprintf( tmp, "%s", str );

	slen = strlen( str );

	if ( slen > len )
	{
		if ( mark_incl )
			ptr = tmp + len - 1;
		
		else
			ptr = tmp + len;

		*ptr++ = '*';

		*ptr = '\0';
	}

	return( copy_str( tmp ) );
}


char *
upper_str( str )
char *str;
{
	char *ptr;
	char *tmp;

	tmp = copy_str( str );

	ptr = tmp;

	while ( *ptr != '\0' )
	{
		if ( *ptr >= 'a' && *ptr <= 'z' )
			*ptr += 'A' - 'a';

		ptr++;
	}

	return( tmp );
}


char *
copy_str( str )
char *str;
{
	char *tmp;

	tmp = (char *) malloc( (STV_SIZE_T) (strlen(str) + 1)
		* sizeof(char) );
	memcheck(tmp,"Copy String");

	strcpy( tmp, str );

	return( tmp );
}


int
append_cmd_str( result, str, length, err, cmd_flag )
char *result;
char *str;
int length;
char *err;
int cmd_flag;
{
	char *strcat();

	char *errstr;

	int elen;
	int rlen;
	int len;
	int rc;
	int i;

	rlen = strlen( result );

	len = rlen + strlen( str ) + 1;

	if ( len <= length )
		strcat( result, str );

	else
	{
		if ( cmd_flag == stvTrue )
		{
			if ( err == NULL )
			{
				printf( "\nError in Append Cmd String: Null Cmd\n\n" );

				return( TCL_ERROR );
			}

			elen = len + strlen( err ) + 5;

			errstr = (char *) malloc( (STV_SIZE_T) elen
				* sizeof(char) );
			memcheck( errstr, "Append Cmd String Err String" );

			sprintf( errstr, "%s { %s }", err, result );

			rc = Tcl_Eval( interp, errstr );

			if ( rc == TCL_ERROR )
				printf( "\nError Executing \"%s\"\n\n", errstr );

			free( errstr );

			if ( strlen( str ) <= length )
				strcpy( result, str );

			else
			{
				printf( "\nError Append Cmd String \"%s\" Too Long\n\n",
					str );

				strcpy( result, "" );
			}

			return( rc );
		}

		else
		{
			if ( err != NULL )
				printf( "\nError in Append Cmd String: %s\n\n", err );

			i = rlen;

			while ( i < length - 4 )
			{
				result[i] = str[ i - rlen ];

				i++;
			}

			sprintf( (char *) (result + (length - 4)), "..." );
		}
	}

	return( TCL_OK );
}


double
distance( x1, y1, x2, y2 )
double x1, y1, x2, y2;
{
	return( sqrt(((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1))) );
}


double
angle_of( x1, y1, x2, y2 )
double x1, y1, x2, y2;
{
	double slope;
	double at;

	/* North is heading angle 0.0 */
	/* y's grow downward from the top of the screen */

	if ( x2 == x1 )
	{
		if ( y2 > y1 )
			at = -1.0 * PI;

		else
			at = 0.0;
	}

	else
	{
		slope = (y1 - y2) / (x2 - x1);

		at = atan( slope );

		if ( x1 > x2 )
			at += PI / 2.0;

		else
			at -= PI / 2.0;
	}

	return( at );
}


int
filecheck( fp, name )
FILE *fp;
char *name;
{
	char msg[1024];
	char tmp[1024];

	if ( fp == NULL )
	{
		sprintf( tmp, "Error Opening File \"%s\"", name );

		sprintf( msg, "setMsg {%s}", tmp );

		Tcl_Eval( interp, msg );

		fprintf( stderr, "\n%s\n\n", tmp );

		return( stvFalse );
	}

	return( stvTrue );
}


void
memcheck( ptr, name )
char *ptr;
char *name;
{
	if ( ptr == NULL )
	{
		fprintf( stderr, "\nError Allocating Memory for \"%s\"\n\n",
			name );

		exit( -1 );
	}
}


int
compare( x, y )
char *x, *y;
{
	while ( *x != '\0' )
	{
		if ( *x != *y )
			return( stvFalse );

		x++; y++;
	}

	return( stvTrue );
}

