
C $Id: fcloud_mpi.f,v 1.3 1999/02/11 22:08:57 kohl Exp $


C
C                       CUMULVS version 1.0:
C       Computational Steering and Interactive Visualization
C                    in Distributed Applications
C            Oak Ridge National Laboratory, Oak Ridge TN.
C       Authors:  James Arthur Kohl and Philip M. Papadopoulos
C                    (C) 1996 All Rights Reserved
C
C                               NOTICE
C
C  Permission to use, copy, modify, and distribute this software and
C  its documentation for any purpose and without fee is hereby granted
C  provided that the above copyright notice appear in all copies and
C  that both the copyright notice and this permission notice appear
C  in supporting documentation.
C
C  Neither the Institution, Oak Ridge National Laboratory, nor the
C  Authors make any representations about the suitability of this
C  software for any purpose.  This software is provided ``as is''
C  without express or implied warranty.
C
C  CUMULVS was funded by the U.S. Department of Energy.
C


C fcloud_mpi.f - A small test program to test the communication
C                  setup with avs.

       program fcloud_mpi
       implicit none
       include 'fpvm3.h'
       include 'fstv.h'
C ---- Parameter declarations
       character*6 FCLOUDGRP
       integer rows, columns
       integer I_DIM, J_DIM, K_DIM, NPROC, DEPTH, STVTAG
       integer NFIELDS, CYCLESZ,CYCLEDEC,NPARAMS,MSGCTXT
       integer RINGMSG,CONFIGMSG
       parameter (FCLOUDGRP='fcloud_mpi')
       parameter(ROWS=2, COLUMNS=2)
       parameter(I_DIM=30, J_DIM=30, K_DIM=120)
       parameter(NPROC=ROWS*COLUMNS, DEPTH=K_DIM)
       parameter(STVTAG=100)
       parameter(MSGCTXT = 101)
       parameter(CONFIGMSG = 1000)
       parameter(RINGMSG = 1001)
       parameter(NFIELDS=2)
       parameter(CYCLESZ=2)
       parameter(NPARAMS=4)
       parameter(CYCLEDEC= CYCLESZ*(1+(J_DIM*COLUMNS)/
     +             (CYCLESZ*COLUMNS)) )
C ----- Variable declarations

       integer mytid, myrow, mycol, myinum
       integer tids(NPROC), info
       integer i,j,k, declare(3), sequence
       real*4  density(I_DIM,J_DIM,K_DIM), value
       real*4  pressure(I_DIM,CYCLEDEC,K_DIM)
       integer rowsize, colsize
       integer centeri,centerj,centerk,globali,globalj,globalk
       integer cglobalj
       integer decompId(NFIELDS), fieldID(NFIELDS)
       integer paramId(NPARAMS)
       real*4  radius, rate, vradius, dist, dist1, vrate
       real*4 params(NPARAMS)
       integer glb(3), gub(3), prank, pshape(2), paddr(2)
       integer axisType(3), axisInfo(3)
       integer restart,np_restart
       integer plist(0:NPROC-1),onum,otid,dest,src
       integer size
       integer gen
C ---- Start of code
       do i=1,I_DIM
          do j = 1, CYCLEDEC
              do k = 1, K_DIM
                   pressure(i,j,k) = 0.0
              enddo
          enddo
       enddo
       call pvmfmytid(mytid)
       print *,'Fcloud started at tid ', mytid
       call pvmfjoingroup(FCLOUDGRP, myinum)
       print *,'Joined Group ', FCLOUDGRP, ' as instance ', myinum

C - Initialize the steering and visualization stuff
C -  NPROC can be any number if it is unknown. stvinit will gets it
C    copy of NPROC set with the call to stvfcpinit, if the task is
C    restarting.

       print *, 'Initializing CUMULVS...'
       call stvfinit( FCLOUDGRP, STVTAG, NPROC, myinum, info)
       print *, '    info = ', info

C - Determine if we are starting from a checkpoint

       print *, 'Initializiing Checkpointing...'
	   call stvfcpinit('fcloud_mpi',MSGCTXT,np_restart,restart)
       print *, '    np_restart = ', np_restart
       print *, '    restart = ', restart

       if (np_restart .gt. 0) then
           print *, 'Cpd gave me', np_restart, 'as # of procs'
       endif

       if (myinum .eq.  0 .and. restart .eq. 0) then
           print *, 'Master Fcloud Task.'
           if (NPROC - 1 .gt. 0) then
                 print *, 'Spawn Remaining Tasks...'
C                call pvmfcatchout(1,info)
                 call pvmfspawn('fcloud_mpi',
C    +               PVMTASKDEBUG,
     +               0,
C    +               PVMTASKHOST+PVMHOSTCOMPL,
C    +               PVMTASKHOST+PVMHOSTCOMPL+PVMTASKDEBUG,
     +                           ' ', NPROC - 1,
     +                         tids, info)
                 if (info .lt. NPROC - 1)  then
                     write(6,*) 'spawn cacked with info = ', info
                 else
                     print *, 'Spawned ', NPROC - 1, ' tasks.'
                     do i = 1, NPROC - 1
                         print *, '    tid=', tids(i)
                     enddo
                 endif
           else
               print *, 'Master Task:  All ', NPROC, ' Tasks Spawned.'
           endif
       else
           print *, 'CP Restart = ', restart
       endif
       myrow = myinum/COLUMNS
       mycol = mod(myinum,COLUMNS)


C -- Record the size of the computational region
       glb(1) = 1
       glb(2) = 1
       glb(3) = 1
       gub(1) = ROWS*I_DIM
       gub(2) = COLUMNS*J_DIM
       gub(3) = K_DIM

C -- Initialize the parameters
       radius = 0
       rate = .2
       vrate = 1.0
       sequence = 0
       params(1) = rate
       params(2) = vrate

       centeri = (I_DIM*ROWS+1)/2
       centerj = (J_DIM*COLUMNS+1)/2
       centerk = (K_DIM+1)/2
       write(6,*) 'center:', centeri, centerj, centerk
C -- Define the steering parameters
       call stvfparamdefine( rate, 'Radius Rate', STVFLOAT,STVVISCP,
     +                          paramId(1))
       call stvfparamdefine( vrate,'Volume Rate',STVFLOAT,STVVISCP,
     +                          paramId(2))
C -- Define checkpointed-only variables
       call stvfparamdefine( radius,'radius',STVFLOAT,STVCPONLY,
     +                          paramId(3))
       call stvfparamdefine( sequence,'sequence',STVINT,STVCPONLY,
     +                          paramId(4))

C - Declare a decomposition of density -- it is block, block, collapsed
       axisType(1) = STVBLOCK
       axisType(2) = STVBLOCK
       axisType(3) = STVCOLLAPSE
       prank = 2
       pshape(1) = ROWS
       pshape(2) = COLUMNS
       call stvfdecompdefine(3,axisType,STVDEFAULTINFO,STVDEFAULTINFO,
     +              glb,gub,prank,pshape,decompId(1))
C - declare the density data  field with the above decomposition
       paddr(1) = myrow
       paddr(2) = mycol
       declare(1) = I_DIM
       declare(2) = J_DIM
       declare(3) = K_DIM
       call stvffielddefine(density,'density',decompId(1),
     +				STVNOFIELDOFFSETS, declare,
     +              STVFLOAT, paddr, STVVISONLY, fieldId(1))
C - Declare a decomposition of pressure -- it is block, cyclic, collapsed
       axisType(1) = STVBLOCK
       axisType(2) = STVCYCLIC
       axisType(3) = STVCOLLAPSE
       axisInfo(1) = STVDEFAULT
       axisInfo(2) = CYCLESZ
       axisInfo(3) = STVDEFAULT
       prank = 2
       pshape(1) = ROWS
       pshape(2) = COLUMNS
       call stvfdecompdefine(3,axisType,axisInfo,axisInfo,glb,gub,
     +                      prank,pshape,decompId(2))
C - declare the pressure field with the above decomposition
       paddr(1) = myrow
       paddr(2) = mycol
       declare(1) = I_DIM
       declare(2) = CYCLEDEC
       declare(3) = K_DIM
       call stvffielddefine(pressure,'pressure',decompId(2),
     +				STVNOFIELDOFFSETS, declare,
     +              STVFLOAT, paddr, STVVISONLY, fieldId(2))

C ---
       colsize = I_DIM * ROWS
       rowsize = J_DIM * COLUMNS

       if (restart .ne. 0) then
          print *, 'I was spawned as a restarted task, load checkpoint'
          call stvfloadfromcp(info)
       endif

C  ------ Jump to here after a load from checkpoint
5       size = 0
        call globlistsetup(FCLOUDGRP,plist,NPROC,myinum,
     >                mytid,MSGCTXT,CONFIGMSG,info)
        if (info .lt. 0) then
            print *,'Got restart while in globlistsetup'
            call stvfloadfromcp(info)
            goto 5
        endif
10     Continue

       radius = radius + rate
       sequence = sequence + 1
       vradius = sequence*vrate
        call stvfgeneration(gen)
        print *,'Sequence, generation ', sequence,gen

       do k = 1, K_DIM
           globalk = k
           do j=1,CYCLEDEC
C Determine from the center of the compgrid, which points should
C Should be colored
                globalj = J + mycol*J_DIM
                cglobalj = (J-1)/CYCLESZ
                cglobalj = (cglobalj*(CYCLESZ*COLUMNS))+MOD(J-1,CYCLESZ)
                cglobalj = cglobalj + mycol*CYCLESZ +1
                do I=1,I_DIM
                   globali = I + myrow*I_DIM
                   dist = (globali - centeri)**2+(globalk - centerk)**2
                   dist1 = sqrt(dist + (cglobalj -centerj)**2 )
                   dist = sqrt(dist + (globalj -centerj)**2 )
                   if (dist .lt. radius) then
                       value = (1.0 - dist/radius)*rate*vradius
                   else
                       value = 0.0
                   endif
               if (j .le. J_DIM) then
                       density(i,j,k) = value
               endif
                   if (dist1 .lt. radius) then
                       value = (1.0 - dist1/radius)*rate*vradius
                   else
                       value = 0.0
                   endif
                   if (cglobalj  .le. (J_DIM*COLUMNS)) then
                       pressure(i,j,k) = value
                   endif
                enddo
           enddo
       enddo

C --    send a message to mod(myinum+1,nproc)
        call stvfuserinitsend(PvmDataDefault,RINGMSG,info)
        call pvmfpack(INTEGER4,myinum,1,1,info)
        call pvmfpack(INTEGER4,mytid,1,1,info)
        dest = myinum + 1
        if (dest .eq. NPROC) dest = 0
        dest = plist(dest)
        src = myinum - 1
        if (src .lt. 0) src = NPROC - 1
        src = plist(src)
        call stvfusersend(dest,MSGCTXT,info)

        call stvfuserrecv(src,RINGMSG,MSGCTXT,info)
        if (info .lt. 0) then
            print *, 'Got restart when trying to read messages'
            print *, 'Info was', info
            call stvfloadfromcp(info)
            goto 5
        endif
        call pvmfunpack(INTEGER4,onum,1,1,info)
        call pvmfunpack(INTEGER4,otid,1,1,info)

C -- send messages to front end viewer
       call stvfsendtofe(info)
       if (MOD(sequence,4) .eq. 0) then
            call stvfcheckpoint(restart)
            if (restart .ne. 0) then
               write(6,*) 'Will reload information from Checkpoint'
               call stvfloadfromcp(info)
               goto 5
            endif
       endif
       call stvfsleep(1)
C      /* End of the computation */
       goto 10
       end

C - Gets the globlist
       subroutine globlistsetup(grp,list,ntask,myinum,
     >                mytid,ctxt,tag,info)
       character*6 grp
       integer list(0:ntask-1), ntask
       integer myinum,mytid,ctxt,tag
       include 'fpvm3.h'
       include 'fstv.h'

       integer size,info,onum,otid

       size = 0
       print *, 'setup:'
       print *, ' grp=',grp
       print *, ' size=',size
       print *, ' ntask=',ntask
       print *, ' myinum=',myinum
       print *, ' mytid=',mytid
       print *, ' ctxt=',ctxt
       print *, ' tag=',tag
       do while (size .lt. ntask)
           call pvmfgsize(grp, size)
           if (size .lt. ntask) call stvfsleep(1)
            print *, 'group size', size
       enddo
        print *, 'group formed'
C --    Bcast a message to everybody in the group
        call stvfuserinitsend(PvmDataDefault,tag,info)
        call pvmfpack(INTEGER4,myinum,1,1,info)
        call pvmfpack(INTEGER4,mytid,1,1,info)
        call pvmfbcast(grp,ctxt,info)

        list(myinum) = mytid
        do i = 1, ntask-1
            call stvfuserrecv(-1,tag,ctxt,info)
            if (info .lt. 0) then
                print *, 'Got restart when trying to read messages'
                print *, 'Info was', info
                return
            endif
            call pvmfunpack(INTEGER4,onum,1,1,info)
            call pvmfunpack(INTEGER4,otid,1,1,info)
            list(onum) = otid
        enddo
        info = 0
        return
        end
