
C $Id: fsimple_mpi.f,v 1.5 1999/03/03 23:14:07 kohl Exp $

C
C                       CUMULVS version 1.0:
C       Computational Steering and Interactive Visualization
C                    in Distributed Applications
C            Oak Ridge National Laboratory, Oak Ridge TN.
C       Authors:  James Arthur Kohl and Philip M. Papadopoulos
C                    (C) 1996 All Rights Reserved
C
C                               NOTICE
C
C  Permission to use, copy, modify, and distribute this software and
C  its documentation for any purpose and without fee is hereby granted
C  provided that the above copyright notice appear in all copies and
C  that both the copyright notice and this permission notice appear
C  in supporting documentation.
C
C  Neither the Institution, Oak Ridge National Laboratory, nor the
C  Authors make any representations about the suitability of this
C  software for any purpose.  This software is provided ``as is''
C  without express or implied warranty.
C
C  CUMULVS was funded by the U.S. Department of Energy.
C

C ---	fsimple_mpi.f - A small test program to test the communication
C ---	setup with CUMULVS.

C ---	This is the MPI version

		program fsimple_mpi

		implicit none
		include 'mpif.h'
		include 'fstv.h'

C --- Parameter declarations

		character*7 FSIMPLEGRP
		integer ROWS, COLUMNS
		integer I_DIM, J_DIM, K_DIM, NPROC, DEPTH, STVTAG
		integer NFIELDS, NPARAMS
		integer RINGMSG

C ---------  Adjustable Parameters --------

C --- The number of processes used is ROWS*COLUMNS

		parameter( ROWS=2, COLUMNS=2 )

C --- The size of the computational region is
C ---  ( ROWS * I_DIM ) X ( COLUMNS * J_DIM ) X ( K_DIM )

		parameter( I_DIM=30, J_DIM=30, K_DIM=120 )

C -----------------------------------------

		parameter( FSIMPLEGRP='fsimple_mpi' )
		parameter( NPROC=ROWS*COLUMNS, DEPTH=K_DIM )
		parameter( STVTAG=100 )
		parameter( RINGMSG=1001 )
		parameter( NFIELDS=1 )
		parameter( NPARAMS=2 )

C --- Variable declarations

		integer axisType(3)
		integer bufid
		integer centeri, centerj, centerk, colsize
		integer declare(3), decompId( NFIELDS ), dest
		integer fieldID( NFIELDS )
		integer glb(3), globali, globalj, globalk, gub(3)
		integer i, idum, info, ierr
		integer j
		integer k
		integer mycol, myinum,  myrow, msrc, mytid
		integer numproc
		integer onum, otid
		integer paddr(2), paramId( NPARAMS ), prank, pshape(2)
		integer rowsize
		integer sequence, src, sendbuf(2)
		real*4 density(I_DIM, J_DIM, K_DIM), dist
		real*4 params( NPARAMS )
		real*4 radius, rate
		real*4 value, vradius, vrate

		integer istatus(MPI_STATUS_SIZE)

C --- Start of code

C ---	MPI Initialization

		call MPI_INIT ( ierr )
		call MPI_COMM_RANK( MPI_COMM_WORLD, myinum, ierr)
		call MPI_COMM_SIZE( MPI_COMM_WORLD, numproc, ierr)

		if ( numproc .ne. NPROC ) then
			print *, 'Number of process should be', NPROC
			print *, 'Exiting.'
			call MPI_FINALIZE( ierr )
			stop
		endif 
		

C --- Determine src and destination for ring message

		dest = myinum + 1
		if ( dest .eq. NPROC ) dest = 0

		src = myinum - 1
		if ( src .lt. 0 ) src = NPROC - 1

C --- Initialize CUMULVS steering and visualization

		call stvfinit( FSIMPLEGRP, STVTAG, NPROC, myinum, info )

C --- Record the size of the computational region

		myrow = myinum / COLUMNS
		mycol = mod( myinum, COLUMNS )

		glb(1) = 1
		glb(2) = 1
		glb(3) = 1

		gub(1) = ROWS * I_DIM
		gub(2) = COLUMNS * J_DIM
		gub(3) = K_DIM

C --- Initialize the parameters

		radius = 0
		rate = .2
		vrate = 1.0
		sequence = 0
		params(1) = rate
		params(2) = vrate

		centeri = ( I_DIM * ROWS + 1 ) / 2
		centerj = ( J_DIM * COLUMNS + 1 ) / 2
		centerk = ( K_DIM + 1 ) / 2

		if ( myinum .eq. 0 ) then
			write(6, *) 'center:', centeri, centerj, centerk
		endif

C --- Define the steering parameters

		call stvfparamdefine( rate, 'Radius Rate', STVFLOAT,
     + 			STVVISONLY, paramId(1) )

		call stvfparamdefine( vrate, 'Volume Rate', STVFLOAT,
     +			STVVISONLY, paramId(2) )

C --- Declare a decomposition of density
C     -- it is block, block, collapsed

		axisType(1) = STVBLOCK
		axisType(2) = STVBLOCK
		axisType(3) = STVCOLLAPSE

		prank = 2
		pshape(1) = ROWS
		pshape(2) = COLUMNS

		call stvfdecompdefine( 3, axisType,
     +		stvDefaultInfo, stvDefaultInfo,
     +		glb, gub, prank, pshape, decompId(1) )

C --- declare the density data field with the above decomposition

		paddr(1) = myrow
		paddr(2) = mycol

		declare(1) = I_DIM
		declare(2) = J_DIM
		declare(3) = K_DIM

		call stvffielddefine( density, 'density', decompId(1),
     +		STVNOFIELDOFFSETS, declare,
     +		STVFLOAT, paddr, STVVISONLY, fieldId(1) )

C ---  Verify Local Field Decomposition

		call stvfdumplocalfieldbounds( fieldId(1), STVVERBOSE, info )

		call stvfdumplocalfieldbounds( fieldId(1), STVTERSE, info )

C ---

		colsize = I_DIM * ROWS
		rowsize = J_DIM * COLUMNS

C ---  Top of main computational loop

10	Continue

		radius = radius + rate
		sequence = sequence + 1
		vradius = sequence*vrate

		if ( myinum .eq. 0 ) then
			print *, 'Sequence ', sequence
		endif

		do k = 1, K_DIM
			globalk = k
			do j=1, J_DIM

C Determine from the center of the compgrid, which points
C should be colored

				globalj = J + mycol*J_DIM
				do I=1, I_DIM
					globali = I + myrow*I_DIM
					dist = ( globali - centeri ) ** 2
     +                  + ( globalk - centerk ) ** 2
					dist = sqrt( dist + ( globalj - centerj ) ** 2 )
					if ( dist .lt. radius ) then
						value = ( 1.0 - dist / radius ) * rate * vradius
					else
						value = 0.0
					endif

					if ( j .le. J_DIM ) then
						density( i, j, k ) = value
					endif
				enddo
			enddo
		enddo

C --- Send a message to mod( myinum + 1, nproc ).
C --- Keep the tasks pseudo-synched

		call stvfsleep(1)
		call MPI_SEND(myinum, 1, MPI_INTEGER, dest, RINGMSG, 
     + 			MPI_COMM_WORLD, istatus)
		call MPI_RECV(onum, 1, MPI_INTEGER, src,
     +                  RINGMSG, MPI_COMM_WORLD, istatus, ierr)


C --- send messages to any attached CUMULVS front-end viewers

		call stvfsendtofe( info )

		if ( info .gt. 0 .and. myinum .eq. 0 ) then
			print *, 'stvfsendtofe returned ', info
		endif
C --- print out message that parameters have been updated by a viewer
		if ( info .gt.  0 .and. myinum .eq. 0 ) then 

			print *,'stv sendToFE returned ', info 

			do i = 1, NPARAMS
				call stvfisparamchanged(paramId(i),info) 
				if ( info .ne. 0 ) then
				 	if (i .eq. 1) then
						print *,
     +					'Parameter "Radius Rate" was changed to ', rate
					endif
				 	if (i .eq. 2) then
						print *,
     +					'Parameter "Volume Rate" was changed to ', vrate
					endif
				endif
			enddo

		endif

C --- This is the end of the Computational loop

		goto 10

		end

