
/* $Id: slview.c,v 1.74 2000/08/03 15:56:31 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#include "slicer.h"
#include "slicer_glob.h"


attach_to_app()
{
	static int hold = 0;

	STV_VIEW_FIELD vfr = (STV_VIEW_FIELD) NULL;
	STV_VIEW_FIELD vfc = (STV_VIEW_FIELD) NULL;

	char tmp[1024];

	int *glb, *gub;

	int dim;
	int i;

	viewer_lock();

	if ( CONNECTED )
	{
		printf( "\nSending FieldHaltAll\n\n" );
		fflush( stdout );

		stv_viewer_send_FieldHaltAll( CURRENT_VFG, VIEWER );

		CONNECTED = stvFalse;
	}

	if ( !INITIALIZED && SLICER_APP != NULL )
	{
		sprintf( tmp,
			"setMsg {Initializing Viewer with Application \"%s\"...}",
			SLICER_APP );
		Tcl_Eval( interp, tmp );

		INITIALIZED =
			( stv_viewer_init( &VIEWER, MYTID, SLICER_APP, stvFalse )
				== stvStatusOk );

		if ( INITIALIZED )
		{
			sprintf( tmp, "setMsg {Connected to Application \"%s\".}",
				SLICER_APP );
			Tcl_Eval( interp, tmp );

			if ( SLICER_FIELD == NULL
				&& SLICER_RADIUS_FIELD == NULL
				&& SLICER_COLOR_FIELD == NULL )
			{
				Tcl_Eval( interp, "make_field_menu" );
			}
		}
	}

	if ( INITIALIZED &&
		( SLICER_FIELD != NULL
			|| SLICER_RADIUS_FIELD != NULL
			|| SLICER_COLOR_FIELD != NULL ) )
	{
		if ( !hold )
		{
			stv_clear_view_field_select( VIEWER );

			if ( SLICER_FIELD != NULL )
			{
				CURRENT_VF = select_field( SLICER_FIELD );

				if ( SLICER_FIELD_NAME != NULL )
					FREE( SLICER_FIELD_NAME );

				SLICER_FIELD_NAME = stv_copy_str( SLICER_FIELD );

				SLICER_FIELD_TYPE = "Field";

				dim = CURRENT_VF->field->decomp->dataDim;

				glb = CURRENT_VF->field->decomp->glb;
				gub = CURRENT_VF->field->decomp->gub;
			}

			else
			{
				if ( SLICER_RADIUS_FIELD != NULL )
					vfr = select_field( SLICER_RADIUS_FIELD );

				if ( SLICER_COLOR_FIELD != NULL )
					vfc = select_field( SLICER_COLOR_FIELD );

				if ( ( vfr != NULL && vfr->field->particle != NULL )
					|| ( vfc != NULL && vfc->field->particle != NULL ) )
				{
					if ( vfr == NULL )
					{
						CURRENT_VF = vfc;

						if ( SLICER_FIELD_NAME != NULL )
							FREE( SLICER_FIELD_NAME );

						SLICER_FIELD_NAME =
							stv_copy_str( SLICER_COLOR_FIELD );

						SLICER_FIELD_TYPE = "Field";
					}
					
					else if ( vfc == NULL )
					{
						CURRENT_VF = vfr;

						if ( SLICER_FIELD_NAME != NULL )
							FREE( SLICER_FIELD_NAME );

						SLICER_FIELD_NAME =
							stv_copy_str( SLICER_RADIUS_FIELD );

						SLICER_FIELD_TYPE = "Field";
					}
					
					else
					{
						if ( strcmp( vfr->field->particle->name,
							vfc->field->particle->name ) )
						{
							printf( "\nError:  %s - %s %s\n\n",
								"Fields from Different Particles",
								"Deferring to Color Field",
								SLICER_COLOR_FIELD );
							fflush( stdout );

							vfr->selected = stvFalse;

							CURRENT_VF = vfc;

							if ( SLICER_FIELD_NAME != NULL )
								FREE( SLICER_FIELD_NAME );

							SLICER_FIELD_NAME =
								stv_copy_str( SLICER_COLOR_FIELD );

							SLICER_FIELD_TYPE = "Field";
						}

						else
						{
							CURRENT_VF = vfc;

							if ( SLICER_FIELD_NAME != NULL )
								FREE( SLICER_FIELD_NAME );

							SLICER_FIELD_NAME = stv_copy_str(
								vfc->field->particle->name );

							SLICER_FIELD_TYPE = "Particle";
						}
					}
				}

				else
					CURRENT_VF = (STV_VIEW_FIELD) NULL;

				if ( CURRENT_VF != NULL )
				{
					dim = CURRENT_VF->field->particle->dataDim;

					glb = CURRENT_VF->field->particle->glb;
					gub = CURRENT_VF->field->particle->gub;
				}
			}

			if ( CURRENT_VF == NULL )
			{
				FREE( SLICER_FIELD_NAME );

				SLICER_FIELD_NAME = (char *) NULL;

				viewer_unlock();

				return( stvFalse );
			}

			if ( !RESTART )
			{
				while ( pop_region( VISREGION ) );

				/* Reset Plane Choice */

				SET_TCL_GLOBAL( interp, "plane_choice", "XY" );

				Tcl_Eval( interp, "layout_main_panel" );

				Tcl_Eval( interp, "set_plane_choice XY" );

				/* Set Max Region */

				for ( i=0 ; i < STV_MAX_DIM ; i++ )
				{
					if ( i < dim )
					{
						STV_REGION_MIN( MAX_REGION, i ) = glb[i];
						STV_REGION_MAX( MAX_REGION, i ) = gub[i];
					}

					else
					{
						STV_REGION_MIN( MAX_REGION, i ) =
							STV_REGION_MAX( MAX_REGION, i ) =
								0;
					}
				}

				/* Set Initial Vis Region */

				for ( i=0 ; i < dim && i < 2 ; i++ )
				{
					STV_REGION_MIN( VISREGION, i ) = glb[i];
					STV_REGION_MAX( VISREGION, i ) = gub[i];

					STV_CELL_OF_DIM( VISCELL, i ) = 1;
				}

				for ( i=((dim < 2) ? dim : 2) ; i < STV_MAX_DIM ; i++ )
				{
					if ( i < dim )
					{
						STV_REGION_MIN( VISREGION, i ) =
							STV_REGION_MAX( VISREGION, i ) =
								glb[i];
					}

					else
					{
						STV_REGION_MIN( VISREGION, i ) =
							STV_REGION_MAX( VISREGION, i ) =
								0;
					}

					STV_CELL_OF_DIM( VISCELL, i ) = 1;
				}

				update_vis_region();
			}

			if ( SLEEP )
			{
				hold++;

				viewer_unlock();

				return( stvFalse );
			}
		}

		if ( !SLEEP )
		{
			/* Request View Field Group */

			sprintf( tmp, "setMsg {Requesting %s \"%s\"...}",
				SLICER_FIELD_TYPE, SLICER_FIELD_NAME );
			Tcl_Eval( interp, tmp );

			CURRENT_VFG = stv_viewer_request_field( VIEWER,
				VISREGION, VISCELL, FREQ, stvFalse, &RESTART );

			if ( CURRENT_VFG == NULL )
			{
				viewer_unlock();

				return( stvFalse );
			}

			FRAME_PENDING = stvFalse;

			CONNECTED = stvTrue;

			REALLOC = stvTrue;

			hold = 0;

			CUMULATIVE_SET = 0;

			if ( CURRENT_VFG->locked == stvAppLocked )
				Tcl_Eval( interp, "lock_label locked" );

			else
				Tcl_Eval( interp, "lock_label unlocked" );

			sprintf( tmp, "setMsg {%s \"%s\" Obtained.}",
				SLICER_FIELD_TYPE, SLICER_FIELD_NAME );
			Tcl_Eval( interp, tmp );
		}
	}

	viewer_unlock();

	return( stvTrue );
}


int
viewer_lock()
{
	VIEWER_LOCK++;

	return( TCL_OK );
}


int
viewer_unlock()
{
	if ( --VIEWER_LOCK > 0 )
		return( TCL_OK );

	if ( VIEWER_LOCK < 0 )
	{
		printf( "Whoa...  Too Many viewer_unlock()'s?!\n" );

		VIEWER_LOCK = 0;
	}

	if ( VIEWER_KICK )
	{
		Tk_DoWhenIdle( viewer_handle, (ClientData) NULL );

		VIEWER_KICK = stvFalse;
	}

	return( TCL_OK );
}


STV_VIEW_FIELD
select_field( name )
char *name;
{
	STV_VIEW_FIELD VF;

	VF = stv_get_view_field_name( name,
		VIEWER->vfields, VIEWER->nvfields );

	if ( VF == NULL )
	{
		printf( "\nError: View Field \"%s\" Not Found\n\n", name );
		fflush( stdout );

		return( (STV_VIEW_FIELD) NULL );
	}

	VF->selected = stvTrue;

	VF->view_type = stvFloat;

	VF->view_storage_order = VF->field->storage_order;

	return( VF );
}


void
viewer_handle( clientData )
ClientData clientData;
{
	STV_VIEW_FIELD_GROUP VFG;

	int lock_check;
	int cc;
	int i;

	/* Check for Viewer Lock */

	if ( VIEWER_LOCK )
	{
		VIEWER_KICK = stvTrue;

		return;
	}

	/* Attach to Region if Not Connected */

	if ( !CONNECTED )
		attach_to_app();

	/* Prevent Nasty Tight Loop... */
	/* If still not connected and no app name, */
	/* stop trying until something changes... */

	if ( !CONNECTED && SLICER_APP == NULL )
	{
		VIEWER_KICK = stvTrue;

		return;
	}

	/* Handle Primary Viewer Functionality (i.e. do stuff) */

	if ( CONNECTED && !XON_KICK )
	{
		if ( !FRAME_PENDING )
		{
			if ( !THRASHING )
			{
				Tcl_Eval( interp,
					"setMsg {Waiting for Data Frame...}" );
			}

			if ( REALLOC )
			{
				if ( !(allocate_data_frame()) )
					return;

				REALLOC = stvFalse;
			}
		}

		/* Save Lock State */

		lock_check = 0;

		/* Collect Data Frame from Application */

		cc = stv_viewer_receive_frame( &VFG, VIEWER, &RESTART,
			stvMsgNoBlock );

		if ( cc == stvStatusIncomplete )
		{
			FRAME_PENDING = stvTrue;

			if ( THRASHING )
			{
				Tcl_Eval( interp,
					"setMsg {Waiting for Data Frame...}" );

				THRASHING = stvFalse;
				THRASH_FACTOR = -1;
			}

			if ( VFG == CURRENT_VFG )
				lock_check++;
		}

		else if ( cc == stvStatusOk || cc == stvStatusBadFrame )
		{
			FRAME_PENDING = stvFalse;

			if ( VFG == CURRENT_VFG )
			{
				if ( !SLEEP || NSTEPS > 0 )
				{
					stv_viewer_send_XON( VFG, VIEWER );

					if ( NSTEPS > 0 )
						NSTEPS--;
				}
				
				else
					XON_KICK = stvTrue;

				if ( cc == stvStatusOk )
					handle_frame( VFG );
				
				else
				{
					Tcl_Eval( interp,
						"setMsg {Bad Data Frame...  Skipped.}" );
				}

				/* Free Field Data */

				stv_reset_Field_Particles( VFG );

				/* Trigger Lock Check */

				lock_check++;
			}

			/* Clean Up Old Field Groups */

			else
				stv_viewer_release_field( &VFG, VIEWER );
		}

		else
		{
			if ( !RESTART )
				RESTART = stvTrue;

			FRAME_PENDING = stvFalse;

			/* Free Field Data */

			for ( i=0 ; i < CURRENT_VFG->nvfields ; i++ )
				stv_free_field_array( CURRENT_VFG, i );

			stv_reset_Field_Particles( CURRENT_VFG );

			REALLOC = stvTrue;

			/* Better Make Sure Appl Knows We Reset */

			stv_viewer_send_FieldHaltAll( CURRENT_VFG, VIEWER );

			do_stv_cleanup();
		}

		/* Handle App Lock */

		if ( lock_check )
		{
			if ( VFG->locked == stvAppLocked )
			{
				if ( cc == stvStatusOk && APP_LOCKED == APP_LOCK_FREE
						&& !SLEEP )
					Tcl_Eval( interp, "sleep_set ON" );

				if ( APP_LOCKED == APP_LOCK_FREE )
					Tcl_Eval( interp, "lock_label locked" );

				else if ( APP_LOCKED == APP_LOCK_LOCKED )
					APP_LOCKED = APP_LOCK_FREE;
			}

			else
			{
				if ( APP_LOCKED == APP_LOCK_FREE )
					Tcl_Eval( interp, "lock_label unlocked" );

				else if ( APP_LOCKED == APP_LOCK_UNLOCKED )
					APP_LOCKED = APP_LOCK_FREE;
			}
		}
	}

	/* Wait Until Something Happens... */

	else
	{
		VIEWER_KICK = stvTrue;

		return;
	}

	/* Set Up Next Handler */

	Tk_DoWhenIdle( viewer_handle, (ClientData) NULL );
}


int
allocate_data_frame()
{
	STV_VIEW_FIELD vf;

	float *saveptr[2];

	int i, j;
	int size;

	/* Free Any Old Diff Fields */

	for ( i=0 ; i < 2 ; i++ )
	{
		saveptr[i] = DIFF_FIELD[i];

		if ( DIFF_FIELD[i] != NULL )
		{
			FREE( DIFF_FIELD[i] );

			DIFF_FIELD[i] = (float *) NULL;
		}
	}

	/* Allocate New Field Storage */

	for ( i=0 ; i < CURRENT_VFG->nvfields ; i++ )
	{
		vf = CURRENT_VFG->view_field_list[i];

		if ( stv_allocate_Field_arrayDecl( CURRENT_VFG, i )
				!= stvStatusOk )
			return( stvFalse );

		if ( vf->field->decomp != NULL )
		{
			/* Don't Free Field Twice! */
			if ( CURRENT_VFG->field_values[i] == saveptr[0]
					|| CURRENT_VFG->field_values[i] == saveptr[1] )
				CURRENT_VFG->field_values[i] = (STV_VALUE) NULL;

			if ( stv_allocate_Field_Data( CURRENT_VFG, i,
				CURRENT_VFG->view_types[i] ) != stvStatusOk )
			{
				return( stvFalse );
			}

			size = 1;

			for ( j=0 ; j < vf->field->decomp->dataDim ; j++ )
				size *= CURRENT_VFG->field_arrayDecls[i][j];

			FIELD_NELEMS = size;

			if ( VIS_MODE == VIS_MODE_DIFF )
			{
				DIFF_FIELD[1] = CURRENT_VFG->field_values[i];

				CURRENT_VFG->field_values[i] = (STV_VALUE) NULL;

				if ( stv_allocate_Field_Data( CURRENT_VFG, i,
					CURRENT_VFG->view_types[i] ) != stvStatusOk )
				{
					printf( "Error: Insufficient Memory for " );
					printf( "Differential Visualization.\n\n" );

					printf( "Reverting to Normal Mode...\n" );

					CURRENT_VFG->field_values[i] = DIFF_FIELD[1];

					DIFF_FIELD[1] = (float *) NULL;

					VIS_MODE = VIS_MODE_NORMAL;
				}

				for ( j=0 ; j < size ; j++ )
					DIFF_FIELD[1][j] = 0.0;

				DIFF_FIELD[0] = CURRENT_VFG->field_values[i];

				DIFF_FIELD_NELEMS = size;

				DIFF_INDEX = 0;
			}
		}
	}

	return( stvTrue );
}


void
handle_frame( VFG )
STV_VIEW_FIELD_GROUP VFG;
{
	static int thrashed = 0;

	STV_VIEW_PARTICLE_DATA VPD;

	STV_VIEW_PARTICLE VPT;

	STV_VIEW_FIELD VF;

	STV_VALUE ptr;

	CIRCLE C, C2;

	CIRCLE newcircles;

	RECT R, R2;

	RECT newrects;

	char color[16];

	char **done_parts;
	int done_already;
	int ndone;

	char *tmp;

	float *dptr;
	int nelems;

	float min, max;
	float pxscale;
	float pyscale;
	float fval;

	int indices[STV_MAX_DIM];

	int slicer_height;
	int slicer_width;
	int radius_index;
	int color_index;
	int has_circles;
	int i, j, k, m;
	int xi, yi, zi;
	int has_rects;
	int ncircles;
	int newsize;
	int border;
	int chgcnt;
	int nrects;
	int xscale;
	int yscale;
	int x1, y1;
	int x2, y2;
	int x, y;
	int huge;
	int redo;
	int dim;
	int r;

	if ( vflag )
		printf( "handle_frame()\n" );

	if ( !(THRASHING = thrashing()) )
	{
		switch ( VIS_MODE )
		{
			case VIS_MODE_NORMAL:
				tmp = "setMsg {Drawing Data Frame...}";
				break;

			case VIS_MODE_DIFF:
				tmp = "setMsg {Drawing Differential Data Frame...}";
				break;

			case VIS_MODE_DIFFSAVE:
				if ( SAVE_FIELD_NELEMS != FIELD_NELEMS )
					tmp =
"setMsg {Drawing Difference to Saved Data Frame (Size Mismatch!)...}";
				else
					tmp =
				"setMsg {Drawing Difference to Saved Data Frame...}";
				break;

			default:
				tmp = "setMsg {Drawing Data Frame...}";
				break;
		}

		Tcl_Eval( interp, tmp );

		thrashed = 0;
	}

	else if ( !thrashed )
	{
		if ( vflag )
			printf( "\tthrashing...\n" );

		Tcl_Eval( interp,
			"setMsg {Processing Data Frames...}" );

		thrashed++;
	}

	/* Get Proper Plane Indices */

	if ( !get_plane_indices( &xi, &yi, &zi ) )
		return;

	/* Allocate Done Particle Array */

	done_parts = (char **) MALLOC( (STV_SIZE_T) VFG->nvfields
			 		* sizeof(char *) );
	memcheck( done_parts, "Done Particles List" );

	for ( i=0 ; i < VFG->nvfields ; i++ )
		done_parts[i] = (char *) NULL;

	ndone = 0;

	/* Go Through All View Fields */

	has_circles = 0;
	has_rects = 0;

	border = 0;

	for ( i=0 ; i < VFG->nvfields ; i++ )
	{
		VF = VFG->view_field_list[i];

		if ( vflag )
			printf( "\tprocessing field %s:\n", VF->field->name );

		/* Contiguous Data */

		if ( VF->field->decomp != NULL )
		{
			if ( vflag )
				printf( "\t\tdecomposed data field.\n" );

			/* Calc Total Num of Rects Needed */

			NX = ( ( STV_REGION_MAX( VFG->visregion, xi )
						- STV_REGION_MIN( VFG->visregion, xi ) + 1 )
					+ ( STV_CELL_OF_DIM( VFG->viscell, xi ) - 1 ) )
				/ STV_CELL_OF_DIM( VFG->viscell, xi );

			NY = ( ( STV_REGION_MAX( VFG->visregion, yi )
						- STV_REGION_MIN( VFG->visregion, yi ) + 1 )
					+ ( STV_CELL_OF_DIM( VFG->viscell, yi ) - 1 ) )
				/ STV_CELL_OF_DIM( VFG->viscell, yi );

			nrects = NX * NY;

			tmp = GET_TCL_GLOBAL( interp, "slicer_height" );
			slicer_height = atoi( tmp );
			slicer_height -= 2 * BORDER_SPACE;

			tmp = GET_TCL_GLOBAL( interp, "slicer_width" );
			slicer_width = atoi( tmp );
			slicer_width -= 2 * BORDER_SPACE;

			xscale = slicer_width / NX;
			yscale = slicer_height / NY;

			SCALE = ( xscale < yscale ) ? xscale : yscale;
			SCALE = SCALE ? SCALE : 1;

			PSCALE = -1.0;

			/* Allocate Rect Array */

			if ( RECTIDS == NULL || RECTIDSIZE < nrects )
			{
				if ( vflag )
					printf( "\t\tallocating rectangles.\n" );

				newsize = RECTIDSIZE ? RECTIDSIZE : 1;

				while ( newsize < nrects )
					newsize *= 2;

				newrects = (RECT) MALLOC( (STV_SIZE_T) newsize
			 		* sizeof(struct rect_struct) );
				memcheck( newrects, "Rectangle IDs List" );

				R = &(newrects[0]);
				R2 = &(RECTIDS[0]);

				for ( j=0 ; j < RECTIDSIZE ; j++ )
				{
					R->x1 = R2->x1;
					R->y1 = R2->y1;
					R->x2 = R2->x2;
					R->y2 = R2->y2;

					R->id = R2->id;

					strcpy( R->color, R2->color );

					R++;  R2++;
				}

				for ( j=RECTIDSIZE ; j < newsize ; j++ )
				{
					R->id = -1;

					*(R->color) = '\0';

					R++;
				}

				if ( RECTIDS != NULL )
					FREE( RECTIDS );

				RECTIDS = newrects;

				RECTIDSIZE = newsize;
			}

			NRECTIDS = nrects;

			/* Automatically Determine Color Map Range */

			if ( AUTO_COLOR_MAP || SHOW_MIN_MAX || SHOW_CUMULATIVE )
			{
				if ( vflag )
					printf( "\t\tdetermining color map range.\n" );

				/* Initialize Local Index */

				for ( j=0 ; j < VF->field->decomp->dataDim ; j++ )
					indices[j] = 0;

				/* Set Data Pointers */

				ptr = VFG->field_values[i];

				switch ( VIS_MODE )
				{
					case VIS_MODE_NORMAL:
						dptr = (float *) NULL;
						break;
				
					case VIS_MODE_DIFF:
						dptr = DIFF_FIELD[ 1 - DIFF_INDEX ];
						nelems = DIFF_FIELD_NELEMS - 1;
						break;
				
					case VIS_MODE_DIFFSAVE:
						dptr = SAVE_FIELD;
						nelems = SAVE_FIELD_NELEMS - 1;
						break;
				}

				/* Determine Min & Max Values */

				j = 0;

				do
				{
					if ( dptr != NULL )
					{
						fval = STV_VALUE_OF( ptr, float ) - (*dptr++);

						if ( j >= nelems )
							dptr = (float *) NULL;
					}

					else
						fval = STV_VALUE_OF( ptr, float );

					if ( !j )
						min = max = fval;

					else
					{
						min = ( fval < min ) ? fval : min;
						max = ( fval > max ) ? fval : max;
					}

					ptr = stv_get_typed_ptr_offset( ptr, 1, stvFloat );

					j++;
				}
				while ( ptr != NULL && j < NRECTIDS
					&& stv_incr_field_index( VF->field->decomp->dataDim,
						indices, VFG->field_arrayDecls[i],
						VFG->visregion, VFG->viscell,
						VFG->view_storage_orders[i], stvFalse ) );

				handle_min_max( min, max );
			}

			/* Draw Rects */

			huge = ( nrects > 10000 );

			do
			{
				if ( vflag )
					printf( "\t\tdrawing rectangles.\n" );

				/* Initialize Local Index */

				for ( j=0 ; j < VF->field->decomp->dataDim ; j++ )
					indices[j] = 0;

				/* Set Data Pointers */

				ptr = VFG->field_values[i];

				switch ( VIS_MODE )
				{
					case VIS_MODE_NORMAL:
						dptr = (float *) NULL;
						break;
				
					case VIS_MODE_DIFF:
						dptr = DIFF_FIELD[ 1 - DIFF_INDEX ];
						nelems = DIFF_FIELD_NELEMS - 1;
						break;
				
					case VIS_MODE_DIFFSAVE:
						dptr = SAVE_FIELD;
						nelems = SAVE_FIELD_NELEMS - 1;
						break;
				}

				/* Draw 'Em */

				R = &(RECTIDS[0]);

				chgcnt = 0;

				redo = 0;

				j = 0;

				do
				{
					x = ( SCALE * indices[xi] ) + BORDER_SPACE;
					y = ( SCALE * indices[yi] ) + BORDER_SPACE;

					if ( dptr != NULL )
					{
						fval = STV_VALUE_OF( ptr, float ) - (*dptr++);

						if ( j >= nelems )
							dptr = (float *) NULL;
					}

					else
						fval = STV_VALUE_OF( ptr, float );

					color_value( color, fval );

					x1 = x;
					y1 = y;
					x2 = x + SCALE - 1;
					y2 = y + SCALE - 1;

					if ( R->id != -1 )
					{
						if ( x1 != R->x1 || y1 != R->y1
							|| x2 != R->x2 || y2 != R->y2 )
						{
							R->x1 = x1;
							R->y1 = y1;
							R->x2 = x2;
							R->y2 = y2;

							RECT_COORDS( interp, SL_C, R->id,
								x1, y1, x2, y2 );

							if ( huge && ++chgcnt > 1000 )
							{
								redo++;

								break;
							}
						}

						if ( strcmp( color, R->color ) )
						{
							strcpy( R->color, color );

							RECT_COLOR( interp, SL_C, R->id,
								color, color );

							if ( huge && ++chgcnt > 1000 )
							{
								redo++;

								break;
							}
						}
					}

					else
					{
						R->x1 = x1;
						R->y1 = y1;
						R->x2 = x2;
						R->y2 = y2;

						strcpy( R->color, color );

						R->id = DRAW_RECT( interp, SL_C,
							x, y, x + SCALE - 1, y + SCALE - 1,
							color, color, 1 );
					}

					R->x = STV_REGION_MIN( VFG->visregion, xi )
						+ ( STV_CELL_OF_DIM( VFG->viscell, xi )
							* indices[xi] );
					R->y = STV_REGION_MIN( VFG->visregion, yi )
						+ ( STV_CELL_OF_DIM( VFG->viscell, yi )
							* indices[yi] );
					R->z = STV_REGION_MIN( VFG->visregion, zi )
						+ ( STV_CELL_OF_DIM( VFG->viscell, zi )
							* indices[zi] );

					R->value = fval;

					ptr = stv_get_typed_ptr_offset( ptr, 1, stvFloat );

					R++;  j++;
				}
				while ( ptr != NULL && j < NRECTIDS
					&& stv_incr_field_index( VF->field->decomp->dataDim,
						indices, VFG->field_arrayDecls[i],
						VFG->visregion, VFG->viscell,
						VFG->view_storage_orders[i], stvFalse ) );

				if ( redo )
				{
					if ( vflag )
						printf(
							"\t\thuge number of rects, redraw all.\n" );

					RECT_DELETE_ALL( interp, SL_C );

					R = &(RECTIDS[0]);

					for ( j=0 ; j < RECTIDSIZE ; j++ )
					{
						if ( R->id != -1 )
						{
							R->id = -1;

							*(R->color) = '\0';
						}

						R++;
					}

					huge = 0;
				}
			}
			while ( redo );

			/* Clean Up Old Rects */

			if ( vflag )
				printf( "\t\tcleaning up any leftover rectangles.\n" );

			while ( j < RECTIDSIZE )
			{
				if ( R->id != -1 )
				{
					DELETE_ITEM( interp, SL_C, R->id );

					R->id = -1;

					*(R->color) = '\0';
				}

				R++;  j++;
			}

			/* Swap Data Storage for Differential Mode */

			if ( VIS_MODE == VIS_MODE_DIFF )
			{
				DIFF_INDEX = 1 - DIFF_INDEX;

				VFG->field_values[i] = DIFF_FIELD[ DIFF_INDEX ];
			}

			has_rects++;
		}

		/* Particle Data */

		else if ( VF->field->particle != NULL )
		{
			if ( vflag )
				printf( "\t\tparticle data field.\n" );

			dim = VF->field->particle->dataDim;

			done_already = 0;

			for ( j=0 ; j < ndone && !done_already ; j++ )
			{
				if ( !strcmp( done_parts[j],
					VF->field->particle->name ) )
				{
					done_already++;
				}
			}

			if ( !done_already )
			{
				if ( vflag )
					printf( "\t\tnot already processed.\n" );

				done_parts[ ndone++ ] = VF->field->particle->name;

				/* Calc Drawing Scale */

				NX = STV_REGION_MAX( VFG->visregion, xi )
					- STV_REGION_MIN( VFG->visregion, xi ) + 1;

				NY = STV_REGION_MAX( VFG->visregion, yi )
					- STV_REGION_MIN( VFG->visregion, yi ) + 1;

				tmp = GET_TCL_GLOBAL( interp, "slicer_height" );
				slicer_height = atoi( tmp );
				slicer_height -= 2 * BORDER_SPACE;

				tmp = GET_TCL_GLOBAL( interp, "slicer_width" );
				slicer_width = atoi( tmp );
				slicer_width -= 2 * BORDER_SPACE;

				pxscale = (float) slicer_width / (float) NX;
				pyscale = (float) slicer_height / (float) NY;

				PSCALE = ( pxscale < pyscale ) ? pxscale : pyscale;

				SCALE = -1;

				/* Draw Vis Region Border */

				BORDERID->x1 = BORDERID->y1 = BORDER_SPACE;

				BORDERID->x2 = (int) ( PSCALE
					* (float) ( STV_REGION_MAX( VFG->visregion, xi )
						- STV_REGION_MIN( VFG->visregion, xi ) ) )
					+ BORDER_SPACE;

				BORDERID->y2 = (int) ( PSCALE
					* (float) ( STV_REGION_MAX( VFG->visregion, yi )
						- STV_REGION_MIN( VFG->visregion, yi ) ) )
					+ BORDER_SPACE;

				if ( BORDERID->id != -1 )
				{
					RECT_COORDS( interp, SL_C, BORDERID->id,
						BORDERID->x1, BORDERID->y1,
						BORDERID->x2, BORDERID->y2 );
				}

				else
				{
					BORDERID->id = DRAW_RECT( interp, SL_C,
						BORDERID->x1, BORDERID->y1,
						BORDERID->x2, BORDERID->y2,
						"", "blue", 1 );
				}

				border++;

				/* Allocate Circle Array */

				VPT = VFG->particle_list;

				ncircles = 0;

				while ( VPT != NULL )
				{
					ncircles += VPT->npdata;

					VPT = VPT->next;
				}

				if ( CIRCLEIDS == NULL || CIRCLEIDSIZE < ncircles )
				{
					if ( vflag )
						printf( "\t\tallocating circles.\n" );

					newsize = CIRCLEIDSIZE ? CIRCLEIDSIZE : 1;

					while ( newsize < ncircles )
						newsize *= 2;

					newcircles = (CIRCLE) MALLOC( (STV_SIZE_T) newsize
			 			* sizeof(struct circle_struct) );
					memcheck( newcircles, "Circle IDs List" );

					C = &(newcircles[0]);
					C2 = &(CIRCLEIDS[0]);

					for ( j=0 ; j < CIRCLEIDSIZE ; j++ )
					{
						C->xc = C2->xc;
						C->yc = C2->yc;
						C->r = C2->r;

						C->id = C2->id;

						strcpy( C->color, C2->color );

						C++;  C2++;
					}

					for ( j=CIRCLEIDSIZE ; j < newsize ; j++ )
					{
						C->id = -1;

						*(C->color) = '\0';

						C++;
					}

					if ( CIRCLEIDS != NULL )
						FREE( CIRCLEIDS );

					CIRCLEIDS = newcircles;

					CIRCLEIDSIZE = newsize;
				}

				NCIRCLEIDS = ncircles;

				/* Automatically Determine Color Map Range */

				if ( AUTO_COLOR_MAP || SHOW_MIN_MAX || SHOW_CUMULATIVE )
				{
					if ( vflag )
						printf( "\t\tdetermining color map range.\n" );

					VPT = VFG->particle_list;
	
					m = 0;
	
					while ( VPT != NULL )
					{
						/* Get Index of Color Value... */
	
						color_index = -1;
	
						for ( j=0 ; j < VPT->nvfields ; j++ )
						{
							if ( SLICER_COLOR_FIELD != NULL
								&& !strcmp(
								(VPT->view_field_list[j])->field->name,
									SLICER_COLOR_FIELD ) )
							{
								color_index = j;
							}
						}
	
						if ( color_index >= 0 )
						{
							VPD = VPT->pdata;
	
							while ( VPD != NULL )
							{
								fval = STV_VALUE_OF(
									VPD->field_values[ color_index ],
									float );

								if ( !m )
									min = max = fval;

								else
								{
									min = ( fval < min ) ? fval : min;
									max = ( fval > max ) ? fval : max;
								}

								VPD = VPD->next;
	
								m++;
							}
						}
	
						VPT = VPT->next;
					}

					if ( m )
						handle_min_max( min, max );
				}

				/* Process Particles */

				VPT = VFG->particle_list;

				C = &(CIRCLEIDS[0]);

				m = 0;

				while ( VPT != NULL )
				{
					if ( vflag )
						printf(
							"\t\tdrawing circles for particle %s.\n",
							VPT->particle->name );

					/* Get Indicies of Radius and/or Color Values */

					radius_index = -1;
					color_index = -1;

					for ( j=0 ; j < VPT->nvfields ; j++ )
					{
						if ( SLICER_RADIUS_FIELD != NULL
							&& !strcmp(
								(VPT->view_field_list[j])->field->name,
								SLICER_RADIUS_FIELD ) )
						{
							radius_index = j;
						}

						else if ( SLICER_COLOR_FIELD != NULL
							&& !strcmp(
								(VPT->view_field_list[j])->field->name,
								SLICER_COLOR_FIELD ) )
						{
							color_index = j;
						}
					}

					VPD = VPT->pdata;

					while ( VPD != NULL )
					{
						C->x = ( (xi < dim) ? VPD->coords[xi] : 0 );
						C->y = ( (yi < dim) ? VPD->coords[yi] : 0 );
						C->z = ( (zi < dim) ? VPD->coords[zi] : 0 );

						C->radius = ( radius_index >= 0 )
							? STV_VALUE_OF(
								VPD->field_values[ radius_index ],
								float )
							: 1.0;

						C->value = ( color_index >= 0 )
							? STV_VALUE_OF(
								VPD->field_values[ color_index ],
								float )
							: 0.0;

						x = (int) ( PSCALE
							* (float) ( C->x
							- STV_REGION_MIN( VFG->visregion, xi ) ) )
							+ BORDER_SPACE;

						y = (int) ( PSCALE
							* (float) ( C->y
							- STV_REGION_MIN( VFG->visregion, yi ) ) )
							+ BORDER_SPACE;

						/* Get Radius Value */

						r = (int) ( PSCALE * C->radius );

						/* Get Color Value */

						color_value( color, C->value );

						/* Draw It */

						if ( C->id != -1 )
						{
							if ( x != C->xc || y != C->yc || r != C->r )
							{
								C->xc = x;
								C->yc = y;
								C->r = r;

								CIRCLE_COORDS( interp, SL_C, C->id,
									x, y, r );
							}

							if ( strcmp( color, C->color ) )
							{
								strcpy( C->color, color );

								CIRCLE_COLOR( interp, SL_C, C->id,
									color, color );
							}
						}

						else
						{
							C->xc = x;
							C->yc = y;
							C->r = r;

							strcpy( C->color, color );

							C->id = DRAW_CIRCLE( interp, SL_C, x, y, r,
								color, color, 1 );
						}

						VPD = VPD->next;

						C++;  m++;
					}

					VPT = VPT->next;
				}

				/* Clean Up Old Circles */

				if ( vflag )
					printf( "\t\tcleaning up any leftover circles.\n" );

				while ( m < CIRCLEIDSIZE )
				{
					if ( C->id != -1 )
					{
						DELETE_ITEM( interp, SL_C, C->id );

						C->id = -1;

						*(C->color) = '\0';
					}

					C++;  m++;
				}
			}

			has_circles++;
		}
	}

	/* Clean Up Leftover Rects & Circles */

	if ( !has_rects )
	{
		R = &(RECTIDS[0]);

		j = 0;

		while ( j < RECTIDSIZE )
		{
			if ( R->id != -1 )
			{
				DELETE_ITEM( interp, SL_C, R->id );

				R->id = -1;

				*(R->color) = '\0';
			}

			R++;  j++;
		}
	}

	if ( !has_circles )
	{
		C = &(CIRCLEIDS[0]);

		m = 0;

		while ( m < CIRCLEIDSIZE )
		{
			if ( C->id != -1 )
			{
				DELETE_ITEM( interp, SL_C, C->id );

				C->id = -1;

				*(C->color) = '\0';
			}

			C++;  m++;
		}
	}

	/* Clean Up Border if Not Needed (No Particles) */

	if ( !border && BORDERID->id != -1 )
	{
		DELETE_ITEM( interp, SL_C, BORDERID->id );

		BORDERID->id = -1;
	}

	/* Free Up Done Particle List */

	for ( i=0 ; i < VFG->nvfields ; i++ )
		done_parts[i] = (char *) NULL;

	FREE( done_parts );

	/* Done */

	if ( !THRASHING )
	{
		if ( CURRENT_VFG->locked == stvAppLocked )
			Tcl_Eval( interp,
				"setMsg {Data Frame Done (Application Locked).}" );

		else
			Tcl_Eval( interp, "setMsg {Data Frame Done.}" );
	}
}


void
handle_min_max( min, max )
float min, max;
{
	static float minsave = 1.0;
	static float maxsave = -1.0;

	static int showsave = -1;

	char minstr[255];
	char maxstr[255];
	char cmd[255];

	int changed = 0;
	int show;

	/* Handle Auto Color Map */

	if ( AUTO_COLOR_MAP )
	{
		if ( COLOR_MIN_VALUE != min || COLOR_MAX_VALUE != max )
		{
			COLOR_MIN_VALUE = min;
			COLOR_MAX_VALUE = max;

			update_color_range();

			safe_eval( interp, "draw_color_map" );
		}
	}

	/* Force Redraw of Min & Max Display If State Has Changed... */

	show = ( ( SHOW_MIN_MAX ) ? 1 : 0 )
		+ ( ( ( SHOW_CUMULATIVE ) ? 1 : 0 ) * 10 );

	if ( show != showsave )
	{
		showsave = show;
		changed++;
	}
	
	/* Handle Min & Max, Cumulative Displays */

	if ( SHOW_CUMULATIVE )
	{
		if ( !CUMULATIVE_SET )
		{
			CUMULATIVE_MIN = min;
			CUMULATIVE_MAX = max;

			CUMULATIVE_SET++;

			changed++;
		}

		else
		{
			if ( min < CUMULATIVE_MIN )
			{
				CUMULATIVE_MIN = min;

				changed++;
			}

			if ( max > CUMULATIVE_MAX )
			{
				CUMULATIVE_MAX = max;

				changed++;
			}
		}
	}

	if ( SHOW_MIN_MAX || SHOW_CUMULATIVE )
	{
		if ( !SHOW_CUMULATIVE )
		{
			if ( min != minsave || max != maxsave )
			{
				sprintf( minstr, "Min: %f", min );
				sprintf( maxstr, "Max: %f", max );

				changed++;
			}
		}
		
		else if ( !SHOW_MIN_MAX )
		{
			if ( changed )
			{
				sprintf( minstr, "Cumulative Min: %f", CUMULATIVE_MIN );
				sprintf( maxstr, "Cumulative Max: %f", CUMULATIVE_MAX );
			}
		}

		else
		{
			if ( changed || min != minsave || max != maxsave )
			{
				sprintf( minstr, "Min: %f / %f", min, CUMULATIVE_MIN );
				sprintf( maxstr, "Max: %f / %f", max, CUMULATIVE_MAX );

				changed++;
			}
		}

		if ( changed )
		{
			sprintf( cmd, "%s.minval configure -text {%s}",
				SL, minstr );
			safe_eval( interp, cmd );

			sprintf( cmd, "%s.maxval configure -text {%s}",
				SL, maxstr );
			safe_eval( interp, cmd );
		}
	}
}


void
redraw_view()
{
	CIRCLE C;

	RECT R;

	char *tmp;

	float oldpscale;
	float pxscale;
	float pyscale;

	int slicer_height;
	int slicer_width;
	int oldscale;
	int xscale;
	int yscale;
	int i;

	/* Get Canvas Dimensions */

	tmp = GET_TCL_GLOBAL( interp, "slicer_height" );
	slicer_height = atoi( tmp );
	slicer_height -= 2 * BORDER_SPACE;

	tmp = GET_TCL_GLOBAL( interp, "slicer_width" );
	slicer_width = atoi( tmp );
	slicer_width -= 2 * BORDER_SPACE;

	/* Re-Scale Rectangles */

	xscale = slicer_width / NX;
	yscale = slicer_height / NY;

	oldscale = SCALE;

	SCALE = ( xscale < yscale ) ? xscale : yscale;
	SCALE = SCALE ? SCALE : 1;

	/* Redraw Rects */

	if ( SCALE != oldscale )
	{
		R = &(RECTIDS[0]);

		for ( i=0 ; i < NRECTIDS ; i++ )
		{
			if ( R->id != -1 )
			{
				R->x1 = ( ((R->x1 - BORDER_SPACE) / oldscale) * SCALE )
						+ BORDER_SPACE;

				R->y1 = ( ((R->y1 - BORDER_SPACE) / oldscale) * SCALE )
						+ BORDER_SPACE;

				R->x2 = R->x1 + SCALE - 1;
				R->y2 = R->y1 + SCALE - 1;

				RECT_COORDS( interp, SL_C, R->id,
					R->x1, R->y1, R->x2, R->y2 );
			}

			R++;
		}
	}

	/* Re-Scale Particles */

	pxscale = (float) slicer_width / (float) NX;
	pyscale = (float) slicer_height / (float) NY;

	oldpscale = PSCALE;

	PSCALE = ( pxscale < pyscale ) ? pxscale : pyscale;

	/* Redraw Circles & Border */

	if ( PSCALE != oldpscale )
	{
		/* Draw Vis Region Border */

		if ( BORDERID->id != -1 )
		{
			BORDERID->x2 = (int)
				( ( (float) (BORDERID->x2 - BORDER_SPACE) / oldpscale )
					* PSCALE )
				+ BORDER_SPACE;

			BORDERID->y2 = (int)
				( ( (float) (BORDERID->y2 - BORDER_SPACE) / oldpscale )
					* PSCALE )
				+ BORDER_SPACE;

			RECT_COORDS( interp, SL_C, BORDERID->id,
				BORDERID->x1, BORDERID->y1,
				BORDERID->x2, BORDERID->y2 );
		}

		/* Process Particles */

		C = &(CIRCLEIDS[0]);

		for ( i=0 ; i < NCIRCLEIDS ; i++ )
		{
			if ( C->id != -1 )
			{
				C->xc = (int)
					( ( (float) (C->xc - BORDER_SPACE) / oldpscale )
						* PSCALE )
					+ BORDER_SPACE;

				C->yc = (int)
					( ( (float) (C->yc - BORDER_SPACE) / oldpscale )
						* PSCALE )
					+ BORDER_SPACE;

				C->r = (int) ( PSCALE * C->radius );

				CIRCLE_COORDS( interp, SL_C, C->id,
					C->xc, C->yc, C->r );
			}

			C++;
		}
	}
}


void
set_current_field()
{
	STV_VIEW_FIELD vf;

	int i, j;
	int size;

	/* Allocate Saved Field Storage & Copy Data */

	for ( i=0 ; i < CURRENT_VFG->nvfields ; i++ )
	{
		vf = CURRENT_VFG->view_field_list[i];

		if ( vf->field->decomp != NULL )
		{
			/* Get Size of Field */

			size = 1;

			for ( j=0 ; j < vf->field->decomp->dataDim ; j++ )
				size *= CURRENT_VFG->field_arrayDecls[i][j];

			/* Allocate Storage */

			if ( SAVE_FIELD != NULL )
				FREE( SAVE_FIELD );

			SAVE_FIELD = (float *) MALLOC( (STV_SIZE_T) size
				* sizeof(float) );
			
			if ( SAVE_FIELD == NULL )
			{
				printf(
					"\nError: Insufficient Memory to Save Field!\n\n" );
			}

			else
			{
				for ( j=0 ; j < size ; j++ )
					SAVE_FIELD[j] =
						((float *) CURRENT_VFG->field_values[i])[j];

				SAVE_FIELD_NELEMS = size;
			}
		}
	}
}


void
free_current_field()
{
	int i;

	if ( SAVE_FIELD != NULL )
	{
		for ( i=0 ; i < SAVE_FIELD_NELEMS ; i++ )
			SAVE_FIELD[i] = -1.0;

		FREE( SAVE_FIELD );

		SAVE_FIELD = (float *) NULL;
	}

	SAVE_FIELD_NELEMS = -1;

	if ( VIS_MODE == VIS_MODE_DIFFSAVE )
		safe_eval( interp, "set_vis_mode diffsave" );
}


void
save_current_field( file )
char *file;
{
	FILE *fp;

	float *fptr;

	int i;

	if ( SAVE_FIELD == NULL )
		set_current_field();

	fp = fopen( file, "w" );

	if ( !filecheck( fp, file ) )
		return;

	fprintf( fp, "%d\n", SAVE_FIELD_NELEMS );

	fptr = SAVE_FIELD;

	for ( i=1 ; i <= SAVE_FIELD_NELEMS ; i++ )
		fprintf( fp, "%f%c", *fptr++, ( !(i % 8) ? '\n' : ' ' ) );

	fclose( fp );
}


void
load_current_field( file )
char *file;
{
	FILE *fp;

	float *fptr;

	int nelems = -1;
	int i;

	fp = fopen( file, "r" );

	if ( !filecheck( fp, file ) )
		return;

	if ( fscanf( fp, "%d", &nelems ) != 1 || nelems <= 0 )
	{
		printf( "%s:  Error Parsing Saved Field File - nelems=%d...\n",
			"load_current_field()", nelems );
		
		return;
	}

	if ( SAVE_FIELD != NULL )
		FREE( SAVE_FIELD );
	
	SAVE_FIELD = (float *) MALLOC( (STV_SIZE_T) nelems
		* sizeof(float) );

	if ( SAVE_FIELD == NULL )
		printf( "\nError: Insufficient Memory to Load Field!\n\n" );

	else
	{
		/* Initialize Saved Field Storage */

		fptr = SAVE_FIELD;

		for ( i=0 ; i < nelems ; i++ )
			*fptr++ = 0.0;

		SAVE_FIELD_NELEMS = nelems;

		/* Read in Field */

		fptr = SAVE_FIELD;

		for ( i=0 ; i < nelems ; i++ )
		{
			if ( fscanf( fp, "%f", fptr++ ) != 1 )
			{
				printf(
				"%s:  Error Parsing Saved Field File (%d of %d)...\n",
					"load_current_field()", i, nelems );

				break;
			}
		}
	}

	fclose( fp );
}


void
color_value( color, value )
char color[16];
float value;
{
	char Rstr[4];
	char Gstr[4];
	char Bstr[4];

	float ratio;

	float R;
	float G;
	float B;

	if ( EQUIV( COLOR_MIN_VALUE, COLOR_MAX_VALUE ) )
	{
		if ( value > COLOR_MIN_VALUE )
		{
			R = 255.0;
			G = 0.0;
			B = 0.0;
		}

		else if ( value < COLOR_MIN_VALUE )
		{
			R = 0.0;
			G = 0.0;
			B = 255.0;
		}

		else
		{
			R = 0.0;
			G = 255.0;
			B = 0.0;
		}
	}

	else
	{
		ratio = ( value - COLOR_MIN_VALUE )
			 / ( COLOR_MAX_VALUE - COLOR_MIN_VALUE );

		B = 255.0 - ( 255.0 * ratio );

		B = ( B < 0 ) ? 0.0 : B;
		B = ( B > 255.0 ) ? 255.0 : B;

		if ( ratio > 0.5 )
			G = 255.0 - ( 255.0 * ( ratio - 0.5 ) * 2.0 );

		else
			G = 255.0 * ratio * 2.0;

		G = ( G < 0 ) ? 0.0 : G;
		G = ( G > 255.0 ) ? 255.0 : G;

		R = 255.0 * ratio;

		R = ( R < 0 ) ? 0.0 : R;
		R = ( R > 255.0 ) ? 255.0 : R;
	}

	if ( (int) R > 0xf )
		sprintf( Rstr, "%2x", (int) R );
	else
		sprintf( Rstr, "0%1x", (int) R );

	if ( (int) G > 0xf )
		sprintf( Gstr, "%2x", (int) G );
	else
		sprintf( Gstr, "0%1x", (int) G );

	if ( (int) B > 0xf )
		sprintf( Bstr, "%2x", (int) B );
	else
		sprintf( Bstr, "0%1x", (int) B );

	sprintf( color, "#%s%s%s", Rstr, Gstr, Bstr );
}

