
/* $Id: stv_particle.c,v 3.12 1999/06/04 17:21:07 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <math.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


/* Particle Routines */


/* Pack SubRegion Particles */

int
stv_pack_SubRegion_Particles( sbuf, V, FG, nparts )
int sbuf;
STV_FE_VIEWERPTR V;
STV_FIELD_GROUP_PTR FG;
int nparts;
{
	STV_PARTICLEPTR *particle_list;

	STV_PARTICLEPTR PT;

	STV_PARTICLE_ID ptid;

	STV_FIELDPTR F;

	STV_VALUE ptr;

	int coords[STV_MAX_DIM];

	int *particle_ids;

	int nfields;
	int i, j;
	int flag;
	int pnum;
	int type;
	int tmp;
	int cc;
	int np;

	/* Verify Valid Number of Particles */

	if ( nparts <= 0 )
	{
		STV_ERROR( "stv_pack_SubRegion_Particles", stvStatusBadParam,
			"Invalid Number of Particles" );

		return( stvStatusBadParam );
	}

	/* Initialize Particle Lists */

	particle_list = (STV_PARTICLEPTR *) MALLOC( (STV_SIZE_T) nparts
		* sizeof( STV_PARTICLEPTR ) );
	STV_MEMCHECK_return( particle_list,
		"Particle Ptr List for Packing" );

	particle_ids = (int *) MALLOC( (STV_SIZE_T) nparts
		* sizeof( int ) );
	STV_MEMCHECK_return( particle_ids, "Particle ID List for Packing" );

	/* Collect Particle Fields */

	np = 0;

	for ( i=0 ; i < FG->nfields ; i++ )
	{
		F = FG->field_list[i];

		/* Snag Particle Field */

		if ( F != NULL && F->particle != NULL && F->data_ready )
		{
			/* Check for Existing Particle ID */

			flag = 0;

			for ( j=0 ; j < np && !flag ; j++ )
			{
				if ( particle_ids[j] == F->particleID )
					flag++;
			}

			if ( !flag )
			{
				particle_list[ np ] = F->particle;

				particle_ids[ np ] = F->particleID;

				np++;
			}
		}
	}

	/* Pack # of Particle Fields */

	stv_pack_int( sbuf, &np, 1, 1 );

	/* Pack Particle Fields */

	for ( i=0 ; i < np ; i++ )
	{
		PT = particle_list[i];

		/* Pack Viewer Field IDs */

		tmp = stvMarkerParticleField;
		stv_pack_int( sbuf, &tmp, 1, 1 );

		/* Pack Particle Name */

		stv_pack_str( sbuf, PT->name );

		/* Count Fields in Particle */

		nfields = 0;

		for ( j=0 ; j < FG->nfields ; j++ )
		{
			F = FG->field_list[j];

			if ( F != NULL && F->particleID == particle_ids[i]
					&& F->data_ready )
				nfields++;
		}

		stv_pack_int( sbuf, &(nfields), 1, 1 );

		/* Pack Field IDs */

		for ( j=0 ; j < FG->nfields ; j++ )
		{
			F = FG->field_list[j];

			if ( F != NULL && F->particleID == particle_ids[i]
					&& F->data_ready )
			{
				stv_pack_int( sbuf, &(FG->field_ids[j]), 1, 1 );

				/* Pack Current Field Time */

				stv_pack_int( sbuf, &(F->time_base), 1, 1 );
				stv_pack_int( sbuf, &(F->time_units), 1, 1 );
			}
		}

		/* Pack Actual Particle Field Data */

		pnum = 1;

		/* Get First Particle */

		(*(PT->get_particle))( pnum, FG->visRegion, &ptid, coords );

		/* Pack Each Particle */

		while ( ptid != NULL )
		{
			/* Pack Particle Marker */

			tmp = stvMarkerParticle;
			stv_pack_int( sbuf, &tmp, 1, 1 );

			/* Pack Particle Coords */

			stv_pack_int( sbuf, &(PT->dataDim), 1, 1 );

			stv_pack_int( sbuf, coords, PT->dataDim, 1 );

			/* Pack Particle Field Data */

			for ( j=0 ; j < FG->nfields ; j++ )
			{
				F = FG->field_list[j];

				/* Pack Each Particle Field Value */

				if ( F != NULL && F->particleID == particle_ids[i]
					&& F->data_ready )
				{
					/* Get Particle's Value for this Field */

					(*(F->get_pfield))( ptid, F->get_pfield_arg, &ptr );

					/* Pack Number of Elements */

					stv_pack_int( sbuf, &(F->nelems), 1, 1 );

					/* Check Type Request & Pack */

					type = FG->field_types[j];

					if ( F->type != type )
					{
						stv_pack_converted_type( sbuf, ptr,
							F->nelems, 1, F->type, type );
					}

					else
					{
						if ( (cc = stv_pack_by_type( sbuf, ptr,
							F->nelems, 1, type )) != stvStatusOk )
						{
							STV_ERROR( "stv_pack_SubRegion_Particles",
								cc, "Typed Pack Failed." );

							FREE( particle_list );
							FREE( particle_ids );

							return( cc );
						}
					}
				}
			}

			/* Increment to Next Particle */

			pnum += STV_CELL_PARTICLE( FG->visCell );

			(*(PT->get_particle))( pnum, FG->visRegion, &ptid, coords );

			/* Pack End of Particle Marker */

			tmp = stvMarkerParticleEnd;
			stv_pack_int( sbuf, &tmp, 1, 1 );
		}

		/* Reset get_particle() */

		(*(PT->get_particle))( -1, FG->visRegion, &ptid, coords );

		/* Pack End of Particle Field Marker */

		tmp = stvMarkerParticleFieldEnd;
		stv_pack_int( sbuf, &tmp, 1, 1 );
	}

	/* Free Particle Lists */

	FREE( particle_list );
	FREE( particle_ids );

	/* A.O.K. */

	return( stvStatusOk );
}


/* Unpack SubRegion Particles */

int
stv_unpack_elems_SubRegion_Particles( rbuf, VFG )
int rbuf;
STV_VIEW_FIELD_GROUP VFG;
{
	STV_VIEW_PARTICLE_DATA VPD;

	STV_VIEW_PARTICLE VPT;

	STV_PARTICLEPTR PT;

	STV_VIEW_FIELD VF;

	char tmpname[1024];

	int *field_id_map;
	int *field_ids;

	int nvfields;
	int index;
	int i, j;
	int flag;
	int type;
	int dim;
	int map;
	int tmp;
	int cc;

	/* Unpack Particle Field Marker */

	stv_unpack_int( rbuf, &tmp, 1, 1 );

	if ( tmp != stvMarkerParticleField )
	{
		sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
			"stvMarkerParticleField", stvMarkerParticleField, tmp );

		STV_ERROR( "stv_unpack_elems_SubRegion_Particles",
			stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Unpack Particle Name */

	stv_unpack_str( rbuf, tmpname );

	/* Unpack Particle Field IDs */

	stv_unpack_int( rbuf, &nvfields, 1, 1 );

	field_ids = (int *) MALLOC( (STV_SIZE_T) nvfields * sizeof( int ) );
	STV_MEMCHECK_return( field_ids, "Particle Field IDs List" );

	for ( i=0 ; i < nvfields ; i++ )
	{
		stv_unpack_int( rbuf, &(field_ids[i]), 1, 1 );

		/* Unpack Current Field Time */

		VF = stv_get_view_field_from_group_id( VFG, field_ids[i],
				(int *) NULL );

		stv_unpack_int( rbuf, &(VF->field->time_base), 1, 1 );
		stv_unpack_int( rbuf, &(VF->field->time_units), 1, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_TIME )
		{
			printf( "Field %s:  time = ", VF->field->name );
			stv_dump_base_time_str( stdout,
					VF->field->time_base, VF->field->time_units,
					stvTrue );
			fflush( stdout );
		}
	}

	/* Look for Existing View Particle */

	VPT = stv_get_view_particle_name( VFG->particle_list, tmpname );

	/* Create New View Particle */

	if ( VPT == NULL )
	{
		VPT = stv_create_view_particle();

		STV_NO_MEM_ERR_return( VPT,
			"stv_unpack_elems_SubRegion_Particles",
			FREE( field_ids ) );

		/* Add to List */

		VPT->next = VFG->particle_list;

		VFG->particle_list = VPT;

		/* Get Particle */

		VPT->particle = stv_get_particle_name_vfg( VFG, tmpname );

		if ( VPT->particle == NULL )
		{
			sprintf( STVERRTXT, "Particle \"%s\" Not Found", tmpname );

			STV_ERROR( "stv_unpack_elems_SubRegion_Particles",
				stvStatusBadParam, STVERRTXT );

			FREE( field_ids );

			return( stvStatusBadParam );
		}

		/* Fill In View Fields */

		VPT->view_field_list = (STV_VIEW_FIELD *) MALLOC(
			(STV_SIZE_T) nvfields * sizeof( STV_VIEW_FIELD ) );
		STV_MEMCHECK_ERR_return( VPT->view_field_list,
			"View Particle Field List", FREE( field_ids ) );

		VPT->view_types = (int *) MALLOC( (STV_SIZE_T) nvfields
			* sizeof( int ) );
		STV_MEMCHECK_ERR_return( VPT->view_types, "View Types List",
			FREE( field_ids ) );

		for ( i=0 ; i < nvfields ; i++ )
		{
			VPT->view_field_list[i] =
				stv_get_view_field_from_group_id( VFG, field_ids[i],
					&index );

			VPT->view_types[i] = VFG->view_types[index];
		}

		VPT->nvfields = nvfields;

		VPT->npdata = 0;
	}

	/* Create Field Index Mapping */
	
	field_id_map = (int *) MALLOC( (STV_SIZE_T) nvfields
		* sizeof( int ) );
	STV_MEMCHECK_ERR_return( field_id_map, "Particle Field Index Map",
		FREE( field_ids ) );

	for ( i=0 ; i < nvfields ; i++ )
	{
		flag = 0;

		for ( j=0 ; j < VPT->nvfields && !flag ; j++ )
		{
			if ( (VPT->view_field_list[j])->field->fieldID
				== field_ids[i] )
			{
				field_id_map[i] = j;

				flag++;
			}
		}

		if ( !flag )
		{
			sprintf( STVERRTXT, "Particle Field ID=%d Not Found",
				field_ids[i] );

			STV_ERROR( "stv_unpack_elems_SubRegion_Particles",
				stvStatusBadParam, STVERRTXT );

			FREE( field_id_map );
			FREE( field_ids );

			return( stvStatusBadParam );
		}
	}

	/* Find Last Particle Data */

	VPD = VPT->pdata;

	if ( VPD != NULL )
	{
		while ( VPD->next != NULL )
			VPD = VPD->next;
	}

	/* Unpack De Data Mon */

	stv_unpack_int( rbuf, &tmp, 1, 1 );

	while ( tmp == stvMarkerParticle )
	{
		/* Get New Particle Data Struct */

		if ( VPD == NULL )
			VPD = VPT->pdata = stv_create_view_particle_data();
		
		else
			VPD = VPD->next = stv_create_view_particle_data();

		STV_NO_MEM_ERR_return( VPD,
			"stv_unpack_elems_SubRegion_Particles",
			( FREE( field_id_map ), FREE( field_ids ) ) );

		(VPT->npdata)++;

		/* Unpack Coord Dimension */

		stv_unpack_int( rbuf, &dim, 1, 1 );

		if ( dim != VPT->particle->dataDim )
		{
			sprintf( STVERRTXT, "Particle Dimension Mismatch %d/%d",
				dim, VPT->particle->dataDim );

			STV_ERROR( "stv_unpack_elems_SubRegion_Particles",
				stvStatusBadParam, STVERRTXT );

			FREE( field_id_map );
			FREE( field_ids );

			return( stvStatusBadParam );
		}

		/* Unpack Coords */

		VPD->coords = (int *) MALLOC( (STV_SIZE_T) dim
			* sizeof( int ) );
		STV_MEMCHECK_ERR_return( VPD->coords, "Particle Data Coords",
			( FREE( field_id_map ), FREE( field_ids ) ) );

		stv_unpack_int( rbuf, VPD->coords, dim, 1 );

		/* Unpack Fields */

		VPD->field_values = (STV_VALUE *) MALLOC( (STV_SIZE_T) nvfields
			* sizeof( STV_VALUE ) );
		STV_MEMCHECK_ERR_return( VPD->field_values,
			"Particle Field Value List",
			( FREE( field_id_map ), FREE( field_ids ) ) );

		VPD->field_nelems = (int *) MALLOC( (STV_SIZE_T) nvfields
			* sizeof( int ) );
		STV_MEMCHECK_ERR_return( VPD->field_nelems,
			"Particle Field Value Nelems List",
			( FREE( field_id_map ), FREE( field_ids ) ) );

		for ( i=0 ; i < nvfields ; i++ )
		{
			map = field_id_map[i];

			type = VPT->view_types[ map ];

			/* Unpack Number of Elements */

			stv_unpack_int( rbuf, &(VPD->field_nelems[ map ]), 1, 1 );

			/* Allocate Field Value Storage */

			VPD->field_values[ map ] = (STV_VALUE) MALLOC(
				(STV_SIZE_T) VPD->field_nelems[ map ]
				* STV_SIZEOF( type ) );
			STV_MEMCHECK_ERR_return( VPD->field_values[ map ],
				"Particle Field Value",
				( FREE( field_id_map ), FREE( field_ids ) ) );

			if ( (cc = stv_unpack_by_type( rbuf,
				VPD->field_values[ map ], VPD->field_nelems[ map ], 1,
				type )) != stvStatusOk )
			{
				STV_ERROR( "stv_unpack_elems_SubRegion_Particles", cc,
					"Typed Unpack Failed." );

				FREE( field_id_map );
				FREE( field_ids );

				return( cc );
			}
		}

		/* Unpack End of Particle Marker */

		stv_unpack_int( rbuf, &tmp, 1, 1 );

		if ( tmp != stvMarkerParticleEnd )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerParticleEnd", stvMarkerParticleEnd, tmp );

			STV_ERROR( "stv_unpack_elems_SubRegion_Particles",
				stvStatusBadParam, STVERRTXT );

			FREE( field_id_map );
			FREE( field_ids );

			return( stvStatusBadParam );
		}

		/* Unpack Next Marker */

		stv_unpack_int( rbuf, &tmp, 1, 1 );
	}

	if ( tmp != stvMarkerParticleFieldEnd )
	{
		sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
			"stvMarkerParticleFieldEnd", stvMarkerParticleFieldEnd,
			tmp );

		STV_ERROR( "stv_unpack_elems_SubRegion_Particles",
			stvStatusBadParam, STVERRTXT );

		FREE( field_id_map );
		FREE( field_ids );

		return( stvStatusBadParam );
	}

	FREE( field_id_map );
	FREE( field_ids );

	return( stvStatusOk );
}


/* Dump Particle Field Routine */

void
stv_dump_particle_field( VPT )
STV_VIEW_PARTICLE VPT;
{
	STV_VIEW_PARTICLE_DATA VPD;

	STV_VIEW_PARTICLE VPTptr;

	char *value_str;

	int i, j;

	VPTptr = VPT;

	while ( VPTptr != NULL )
	{
		printf( "\nFor Particle \"%s\":\n\n", VPTptr->particle->name );

		VPD = VPTptr->pdata;

		while ( VPD != NULL )
		{
			printf( "[" );

			for ( i=0 ; i < VPTptr->particle->dataDim - 1 ; i++ )
				printf( "%d,", VPD->coords[i] );
			
			printf( "%d]",
				VPD->coords[ VPTptr->particle->dataDim - 1 ] );

			for ( i=0 ; i < VPTptr->nvfields ; i++ )
			{
				printf( " %s%s",
					(VPTptr->view_field_list[i])->field->name,
					( VPD->field_nelems[i] == 1 ) ? "=" : " = [ " );

				for ( j=0 ; j < VPD->field_nelems[i] ; j++ )
				{
					value_str = stv_str_of_typed_data(
						((char *) VPD->field_values[i])
							+ (j * STV_SIZEOF( VPTptr->view_types[i] )),
						VPTptr->view_types[i] );

					printf( "%s%s", value_str,
						( VPD->field_nelems[i] == 1 ) ? "" : " " );

					FREE( value_str );
				}

				if ( VPD->field_nelems[i] != 1 )
					printf( "]" );
			}

			printf( "\n" );

			VPD = VPD->next;
		}

		printf( "\n" );
		fflush( stdout );

		VPTptr = VPTptr->next;
	}
}

