
/* $Id: stvmac.h,v 3.10 1999/06/18 21:26:14 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* Version define stuff */

#define STV_VERSION "1.1.0"

#define STV_VERSION_NAME "Jake"


/* Various Macros stv user library. */


/******* ERROR MACROS ************/

#define STV_ERROR( _caller, _code, _msg ) \
	fprintf( stderr, "%s(%s): %s\n", _caller, \
		STVERRNAMES[ -(_code) ], _msg )

#define STV_NO_MEM_return( _ptr, _caller ) \
\
	if ( (_ptr) == NULL ) \
	{ \
		STV_ERROR( _caller, stvStatusNoMem, \
			"Couldn't Allocate Memory" ); \
\
		return( stvStatusNoMem ); \
	} \
\

#define STV_NO_MEM_TYPE_return( _ptr, _caller, _type ) \
\
	if ( (_ptr) == NULL ) \
	{ \
		STV_ERROR( _caller, stvStatusNoMem, \
			"Couldn't Allocate Memory" ); \
\
		return( (_type) NULL ); \
	} \
\

#define STV_NO_MEM_ERR_return( _ptr, _caller, _err ) \
\
	if ( (_ptr) == NULL ) \
	{ \
		STV_ERROR( _caller, stvStatusNoMem, \
			"Couldn't Allocate Memory" ); \
\
		_err; \
\
		return( stvStatusNoMem ); \
	} \
\


/******* MEMORY MACROS ************/

#ifdef WIN32
#define BCOPY(_dest,_src,_size) memcpy(_dest,_src,_size)
#else
#define BCOPY(_dest,_src,_size) bcopy(_src,_dest,_size)
#endif

#define BZERO(_src,_len) bzero(_src,_len)


#ifdef STV_MEM_DEBUG

/******* Debug Versions ************/

extern void *STV_TMP_PTR;

#define MALLOC( _d ) \
	( printf( "malloc size = %u\n", _d ), \
		STV_TMP_PTR = (void *) malloc( _d ), \
		printf( "\tmem returned = 0x%x\n", STV_TMP_PTR ), \
		fflush( stdout ), \
		STV_TMP_PTR )

#define CALLOC( _d, _caller ) \
	( printf( "malloc size = %u (%s)\n", _d, _caller ), \
		STV_TMP_PTR = (void *) calloc( _d, sizeof( char ) ), \
		printf( "\tmem returned = 0x%x\n", STV_TMP_PTR ), \
		fflush( stdout ), \
		STV_TMP_PTR )

#define FREE( _c ) \
	( printf( "free ptr = 0x%x\n", _c ), fflush( stdout ), free( _c ) )

#else

/****** Non-Debug Versions ********/

#define MALLOC( _d )    malloc( _d )

#define CALLOC( _d, _caller )   calloc( _d, sizeof( char ) )

#define FREE( _c )    free( _c )

#endif


/**********************************/

#define CPALLOC( _dest, _src, _n, _t, _caller ) \
	{ (_dest) = (_t *) CALLOC( sizeof( int ) * (STV_SIZE_T) (_n), \
			_caller ); \
		if ( (_dest) != (_t *) NULL ) \
			BCOPY( _dest, _src, (STV_SIZE_T) (_n) * sizeof( _t ) ); }


/*** stv_memcheck() Handler Macros ***/

#define STV_MEMCHECK( _ptr, _str, _err ) \
\
	if ( stv_memcheck( (void *) _ptr, _str ) != stvStatusOk ) \
		_err; \
\

#define STV_MEMCHECK_return( _ptr, _str ) \
\
	if ( stv_memcheck( (void *) _ptr, _str ) != stvStatusOk ) \
		return( stvStatusNoMem ); \
\

#define STV_MEMCHECK_ERR_return( _ptr, _str, _err ) \
\
	if ( stv_memcheck( (void *) _ptr, _str ) != stvStatusOk ) \
	{ \
		_err; \
		return( stvStatusNoMem ); \
	} \
\


/******** LINKED LIST MACROS ***********
*
*     o = original (existing) element
*     n = new element to be added
*     e = existing element
*     prv = structure element that points to previous element in list
*     nxt = structure element that points to next element in list
*/


/* put n after o, o must be a non-null */

#define	LISTPUTAFTER(_o,_n,_prv,_nxt) \
	{ (_n)->_nxt=(_o)->_nxt; (_n)->_prv=_o; \
		if ( (_o)->_nxt != NULL ) \
			(_o)->_nxt->_prv =_n; \
		(_o)->_nxt=_n; }


/* put n before o, o must be non-null */

#define	LISTPUTBEFORE(_o,_n,_prv,_nxt) \
	{ (_n)->_prv=(_o)->_prv; (_n)->_nxt=_o; \
		if ( (_o)->_prv != NULL ) \
			(_o)->_prv->_nxt=_n; \
		(_o)->_prv=_n; }


/* delete e from list, e must me non-null */

#define	LISTDELETE(e,_prv,_nxt) \
	{ if ( (e)->_nxt != NULL) \
		(e)->_nxt->_prv=(e)->_prv; \
	(e)->_prv->_nxt=(e)->_nxt; \
	(e)->_prv=(e)->_nxt=0; }

#define FREELIST(_hd,_prv,_nxt) \
        {  while (_hd) \
            {  if ((_hd)->_nxt)\
                 (_hd) = (_hd)->_nxt; \
              else \
                  break; } \
           while ( (_hd) ) \
           {  if ((_hd)->_prv) \
              { (_hd) = (_hd)->_prv; \
                 FREE((_hd)->_nxt); \
              } else \
                  break; } \
           if (_hd) FREE(_hd); \
           (_hd) = NULL; }

#define FREELIST_ANDDATA(_hd,_data,_prv,_nxt) \
        {  while (_hd) \
           {  if ((_hd)->_nxt) {\
                  if ((_hd)->_data != NULL){ \
                       FREE((_hd)->_data); (_hd)->_data = NULL; }\
                 (_hd) = (_hd)->_nxt;  }\
              else \
                  break; } \
           while ( (_hd) ) \
           {  if ((_hd)->_prv) \
              { (_hd) = (_hd)->_prv; \
                 FREE((_hd)->_nxt); \
              } else \
                  break; } \
           if (_hd) FREE(_hd); \
           (_hd) = NULL; }


/********** VISUALIZATION REGION MACROS **********/

#define CHECK_RANGE( _val, _min, _max ) \
	( ( _val < _min ) ? (_min) : \
		( ( _val > _max ) ? (_max) : (_val) ) )


/********** STV_SIZEOF MACRO ***********/

#define STV_SIZEOF( _t ) \
	( (_t) == stvByte ? sizeof(char) : \
	 ( (_t) == stvStr ? sizeof(char *) : \
	  ( (_t) == stvShort ? sizeof(short) : \
	   ( (_t) == stvInt ? sizeof(int) : \
	    ( (_t) == stvLong ? sizeof(long) : \
	     ( (_t) == stvFloat ? sizeof(float) : \
	      ( (_t) == stvDouble ? sizeof(double) : \
	       ( (_t) == stvCplx ? (STV_SIZE_T) 2*sizeof(float) : \
	        ( (_t) == stvDcplx ? (STV_SIZE_T) 2*sizeof(double) : \
	         ( (_t) == stvUshort ? sizeof(unsigned short) : \
	          ( (_t) == stvUint ? sizeof(unsigned int) : \
	           ( (_t) == stvUlong ? sizeof(unsigned long) : \
	            (STV_SIZE_T) 0 ) ) ) ) ) ) ) ) ) ) ) )

